import logging
import importlib
import os
from collections import defaultdict

from PySide6.QtCore import QObject, Signal
from PySide6.QtWidgets import QFileDialog
from jal.constants import Setup
from jal.db.helpers import get_app_path
from jal.db.settings import JalSettings, FolderFor
from jal.data_import.statement import Statement_ImportError, Statement_Capabilities


# ----------------------------------------------------------------------------------------------------------------------
class Statements(QObject):
    load_completed = Signal(int, defaultdict)
    load_failed = Signal()

    def __init__(self, parent):
        super().__init__()
        self.parent = parent

        self.items = []
        self.loadStatementsList()

    def loadStatementsList(self):
        statements_folder = get_app_path() + Setup.IMPORT_PATH + os.sep + Setup.STATEMENT_PATH
        statement_modules = [filename[:-3] for filename in os.listdir(statements_folder) if filename.endswith(".py")]
        for module_name in statement_modules:
            logging.debug(f"Trying to load statement module: {module_name}")
            try:
                module = importlib.import_module(f"jal.data_import.broker_statements.{module_name}")
            except ImportError:
                logging.error(self.tr("Statement module can't be imported: ") + module_name)
                continue
            try:
                statement_class_name = getattr(module, "JAL_STATEMENT_CLASS")
            except AttributeError:
                continue
            try:
                class_instance = getattr(module, statement_class_name)
            except AttributeError:
                logging.error(self.tr("Statement class can't be loaded: ") + statement_class_name)
                continue
            statement = class_instance()
            self.items.append({
                'name': statement.name,
                'module': module,
                'loader_class': statement_class_name,
                'icon': statement.icon_name,
                'filename_filter': statement.filename_filter
            })
            logging.debug(f"Class '{statement_class_name}' providing '{statement.name}' statement has been loaded")
        self.items = sorted(self.items, key=lambda item: item['name'])

    # method is called directly from menu, so it contains QAction that was triggered
    def load(self, action):
        statement_loader = self.items[action.data()]
        folder = JalSettings().getRecentFolder(FolderFor.Statement, '.')
        statement_files, active_filter = QFileDialog.getOpenFileNames(None, self.tr("Select statement files to import"),
                                                                      folder, statement_loader['filename_filter'])
        if not statement_files:
            return
        JalSettings().setRecentFolder(FolderFor.Statement, statement_files[0])

        module = statement_loader['module']
        class_instance = getattr(module, statement_loader['loader_class'])
        if len(statement_files) > 1:
            if not Statement_Capabilities.MULTIPLE_LOAD in class_instance.capabilities():
                logging.warning(statement_loader['name'] +
                                self.tr(" - module doesn't support multiple statements load."))
                return
            statement_files = class_instance.order_statements(statement_files)
        if not statement_files:
            return
        for statement_file in statement_files:
            statement = class_instance()
            try:
                statement.load(statement_file)
                statement.validate_format()
                statement.match_db_ids()
                totals = statement.import_into_db()
            except Statement_ImportError as e:
                logging.error(self.tr("Import failed: ") + str(e))
                self.load_failed.emit()
                return
        self.load_completed.emit(statement.period()[1], totals)



from decimal import Decimal
from jal.constants import BookAccount
from jal.db.db import JalDB
from jal.db.asset import JalAsset
from jal.db.operations import IncomeSpending


class JalCategory(JalDB):
    def __init__(self, id: int = 0):
        super().__init__()
        self._id = id
        self._data = self._read("SELECT pid, name FROM categories WHERE id=:category_id",
                                [(":category_id", self._id)], named=True)
        self._pid = self._data['pid'] if self._data is not None else 0
        self._name = self._data['name'] if self._data is not None else None

    def id(self) -> int:
        return self._id

    def parent_id(self) -> int:
        return self._pid

    def name(self) -> str:
        return self._name

    # Returns a list of JalCategory objects that represent child categories of the current category
    def get_child_categories(self) -> list:
        children = []
        query = self._exec("SELECT id FROM categories WHERE pid=:category_id", [(":category_id", self._id)])
        while query.next():
            children.append(JalCategory(self._read_record(query)))
        return children

    # Calculates overall turnover in ledger for the category between begin and end timestamps in given currency
    # (conversion rate is used for the day of operation)
    def get_turnover(self, begin: int, end: int, output_currency_id: int) -> Decimal:
        turnover = Decimal('0')
        query = self._exec("SELECT l.timestamp, l.amount, a.currency_id FROM ledger l "
                           "LEFT JOIN accounts AS a ON l.account_id=a.id "
                           "WHERE (l.book_account=:book_costs OR l.book_account=:book_incomes) "
                           "AND l.timestamp>=:begin AND l.timestamp<=:end AND l.category_id=:category_id",
                           [(":book_costs", BookAccount.Costs), (":book_incomes", BookAccount.Incomes),
                            (":begin", begin), (":end", end), (":category_id", self._id)])
        while query.next():
            timestamp, amount, currency_id = self._read_record(query, cast=[int, Decimal, int])
            if currency_id == output_currency_id:
                rate = Decimal('1')
            else:
                rate = JalAsset(currency_id).quote(timestamp, output_currency_id)[1]
            turnover += amount * rate
        return -turnover

    def add_or_update_mapped_name(self, name: str) -> None:
        _ = self._exec("INSERT OR REPLACE INTO map_category (value, mapped_to) "
                       "VALUES (:item_name, :category_id)",
                       [(":item_name", name), (":category_id", self._id)], commit=True)

    # Returns a list of all names that were mapped to some category in for of {"value", "mapped_to"}
    @classmethod
    def get_mapped_names(cls) -> list:
        mapped_list = []
        query = cls._exec("SELECT value, mapped_to FROM map_category")
        while query.next():
            mapped_list.append(cls._read_record(query, named=True))
        return mapped_list

    # Returns a list of operations that include this category
    def get_operations(self, begin: int, end: int) -> list:
        operations = []
        query = self._exec("SELECT DISTINCT a.id FROM actions a LEFT JOIN action_details d ON a.id=d.pid "
                           "WHERE d.category_id=:category AND a.timestamp>=:begin AND a.timestamp<:end",
                           [(":category", self._id), (":begin", begin), (":end", end)])
        while query.next():
            operations.append(IncomeSpending(self._read_record(query, cast=[int])))
        return operations

    def replace_with(self, new_id):
        self._exec("UPDATE action_details SET category_id=:new_id WHERE category_id=:old_id",
                   [(":new_id", new_id), (":old_id", self._id)])
        self._exec("UPDATE map_category SET mapped_to=:new_id WHERE mapped_to=:old_id",
                   [(":new_id", new_id), (":old_id", self._id)])
        self._exec("DELETE FROM categories WHERE id=:old_id", [(":old_id", self._id)], commit=True)
        self._id = 0



from datetime import datetime, timezone
from decimal import Decimal

from jal.db.account import JalAccount
from jal.db.asset import JalAsset
from jal.data_export.dlsg import DLSG


class TaxesFlowRus:
    def __init__(self):
        self.year_begin = 0
        self.year_end = 0
        self.flows = {}

    def get_account_values(self, account, date):
        values = []
        assets = account.assets_list(date)
        assets_value = Decimal('0')
        for asset_data in assets:
            assets_value += asset_data['amount'] * asset_data['asset'].quote(date, account.currency())[1]
        if assets_value != Decimal('0'):
            values.append({'account': account.number(), 'currency': JalAsset(account.currency()).symbol(),
                           'is_currency': False, 'value': assets_value})
        money = account.get_asset_amount(date, account.currency())
        if money != Decimal('0'):
            values.append({'account': account.number(),'currency': JalAsset(account.currency()).symbol(),
                           'is_currency': True, 'value': money})
        return values

    def prepare_flow_report(self, year):
        self.flows = {}
        self.year_begin = int(datetime.strptime(f"{year}", "%Y").replace(tzinfo=timezone.utc).timestamp())
        self.year_end = int(datetime.strptime(f"{year + 1}", "%Y").replace(tzinfo=timezone.utc).timestamp())

        accounts = JalAccount.get_all_accounts(active_only=False)
        # collect data for start and end of the period
        values_begin = []
        values_end = []
        for account in accounts:
            if account.country().code() == 'xx' or account.country().code() == 'ru':
                continue
            values_begin += self.get_account_values(account, self.year_begin)
            values_end += self.get_account_values(account, self.year_end)
        values_begin = sorted(values_begin, key=lambda x: (x['account'], x['is_currency'], x['currency']))
        values_end = sorted(values_end, key=lambda x: (x['account'], x['is_currency'], x['currency']))
        for item in values_begin:
            self.append_flow_values(item, "begin")
        for item in values_end:
            self.append_flow_values(item, "end")

        # collect money and assets ins/outs
        flows = [
            {'type': JalAccount.MONEY_FLOW, 'direction': 'in'},
            {'type': JalAccount.MONEY_FLOW, 'direction': 'out'},
            {'type': JalAccount.ASSETS_FLOW, 'direction': 'in'},
            {'type': JalAccount.ASSETS_FLOW, 'direction': 'out'}
        ]
        for account in accounts:
            if account.country().code() == 'xx' or account.country().code() == 'ru':
                continue
            for flow in flows:
                value = account.get_flow(self.year_begin, self.year_end, flow['type'], flow['direction'])
                if value != Decimal('0'):
                    values = {'account': account.number(), 'currency': JalAsset(account.currency()).symbol(),
                              'is_currency': (flow['type'] == JalAccount.MONEY_FLOW), 'value': value}
                    self.append_flow_values(values, flow['direction'])

        report = []
        for account in self.flows:
            for currency in self.flows[account]:
                record = self.flows[account][currency]
                row = {'report_template': "account_lines",
                       'account': account,
                       'currency': f"{currency} ({record['code']})",
                       'money': "Денежные средства",
                       'assets': "Финансовые активы"}
                for dtype in ['money', 'assets']:
                    for key in ['begin', 'in', 'out', 'end']:
                        param = f"{dtype}_{key}"
                        try:
                            row[param] = record[dtype][key] / Decimal('1000')
                        except KeyError:
                            row[param] = Decimal('0')
                report.append(row)
        return report

    # values are dictionary with keys {'account', 'currency', 'is_currency', 'value'}
    # this method puts it into self.flows array that has another structure:
    # { account: {currency: {0: {'value+suffix': X.XX}}, 1: {'value+suffix': SUM(X.XX)} } } }
    def append_flow_values(self, values, name):
        account = values['account']
        try:
            f_account = self.flows[account]
        except KeyError:
            f_account = self.flows[account] = {}
        currency = values['currency']
        try:
            f_currency = f_account[currency]
        except KeyError:
            try:
                currency_code = DLSG.currencies[currency]['code']
            except KeyError:
                currency_code = 'XXX'  # currency code isn't known
            f_currency = f_account[currency] = {'money': {}, 'assets': {}, 'code': currency_code}
        if values['is_currency'] == 0:
            try:
                f_currency['assets'][name] += values['value']
            except KeyError:
                f_currency['assets'][name] = values['value']
        else:                 # addition isn't required below as there should be only one value for money
            f_currency['money'][name] = values['value']



import logging
import re
from datetime import datetime, timezone

from jal.constants import PredefinedCategory
from jal.data_import.statement import FOF, Statement_ImportError
from jal.data_import.statement_xls import StatementXLS
from jal.db.asset import JalAsset

JAL_STATEMENT_CLASS = "StatementJ2T"


# ----------------------------------------------------------------------------------------------------------------------
class StatementJ2T(StatementXLS):
    PeriodPattern = (1, 0, r"Report by Client Accounts over a Period from (?P<S>\d\d\.\d\d\.\d\d\d\d) to (?P<E>\d\d\.\d\d\.\d\d\d\d)")
    AccountPattern = (5, 4, r"(?P<ACCOUNT>\S*)\\.*")
    HeaderCol = 1
    money_section = "ДС на конец периода"
    money_columns = {
        "settled_end": "Сумма",
        "bonus": "Бонус",
        "currency": "Валюта"
    }
    asset_section = "Открытые позиции на конец периода"
    asset_columns = {
        "name": "Instrument description \(Название инструмента\)",
        "isin": "ISIN",
        "symbol": "Symbol \(Символ\)",
        "currency": "Instrument сurrency \(валюта инструмента\)"
    }

    def __init__(self):
        super().__init__()
        self.name = self.tr("Just2Trade")
        self.icon_name = "j2t.png"
        self.filename_filter = self.tr("Just2Trade statement (*.xlsx)")
        self.asset_withdrawal = []

    def _load_currencies(self):
        pass

    def _load_money(self):
        cnt = 0
        row, headers = self.find_section_start(self.money_section, self.money_columns)
        if row < 0:
            return False
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':
                break
            self._update_account_balance(self._statement[headers['currency']][row], 0, 0,
                                         self._statement[headers['settled_end']][row])
            cnt += 1
            row += 1
        logging.info(self.tr("Cash balances loaded: ") + f"{cnt}")

    def _load_assets(self):
        cnt = 0
        row, headers = self.find_section_start(self.asset_section, self.asset_columns)
        if row < 0:
            return False
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':
                break
            asset_name = self._statement[headers['name']][row]
            if asset_name.endswith('*'):    # strip ending star if required
                asset_name = asset_name[:-1]
            currency_code = self.currency_id('USD')      # FIXME put account currency here
            if not self._statement[headers['isin']][row]:
                self.asset_id({'type': FOF.ASSET_CRYPTO, 'symbol': asset_name,
                               'name': asset_name, 'currency': currency_code})
            elif self._statement[headers['symbol']][row]:
                self.asset_id({'type': FOF.ASSET_STOCK, 'isin': self._statement[headers['isin']][row],
                               'symbol': self._statement[headers['symbol']][row],
                               'name': asset_name, 'currency': currency_code})
            else:
                self.asset_id({'type': FOF.ASSET_STOCK, 'isin': self._statement[headers['isin']][row],
                               'name': asset_name, 'currency': currency_code, 'search_online': "MOEX"})
            cnt += 1
            row += 1
        logging.info(self.tr("Securities loaded: ") + f"{cnt}")

    def _load_deals(self):
        self._load_stock_deals()
        self._load_crypto_deals()

    def _load_stock_deals(self):
        cnt = 0
        columns = {
            "number": "Номер сделки",
            "timestamp": "Дата сделки",
            "settlement": "Дата расчетов",
            "asset_name": "Наименование",
            "isin": "ISIN",
            "asset": "Symbol",
            "B/S": "Тип сделки",
            "qty": "Кол-во",
            "fee": "Комиссия брокера в валюте счета",
            "amount": "Итого в валюте счета",
            "fee_ex": "Прочие комиссии в валюте счета"
        }

        row, headers = self.find_section_start("Сделки с Акциями, Паями, Депозитарными расписками ", columns,
                                               header_height=3)
        if row < 0:
            return
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':
                break
            try:
                ts_string = self._statement[headers['timestamp']][row]
                timestamp = int(datetime.strptime(ts_string,
                                                  "%d.%m.%Y %H:%M:%S, %Z%z").replace(tzinfo=timezone.utc).timestamp())
            except ValueError:  # Skip 'Итого' and similar lines
                row += 1
                continue
            deal_number = str(self._statement[headers['number']][row])
            asset_id = self.asset_id({'type': FOF.ASSET_STOCK, 'isin': self._statement[headers['isin']][row],
                                      'symbol': self._statement[headers['asset']][row],
                                      'name': self._statement[headers['asset_name']][row],
                                      'currency': self.currency_id('USD')})  # FIXME - replace hardcoded 'USD'
            if self._statement[headers['B/S']][row] == 'Купля':
                qty = self._statement[headers['qty']][row]
            elif self._statement[headers['B/S']][row] == 'Продажа':
                qty = -self._statement[headers['qty']][row]
            else:
                row += 1
                logging.warning(self.tr("Unknown trade type: ") + self._statement[headers['B/S']][row])
                continue
            # Collect fees
            fee = self._statement[headers['fee']][row] if self._statement[headers['fee']][row] else 0.0
            if self._statement[headers['fee_ex']][row]:
                fee += self._statement[headers['fee_ex']][row]
            # Calculate price in account currency
            amount = self._statement[headers['amount']][row]
            price = -(amount + fee) / qty
            assert price > 0.0
            # Settlement is stored as date in Excel report file
            settlement = int(self._statement[headers['settlement']][row].replace(tzinfo=timezone.utc).timestamp())
            account_id = self._find_account_id(self._account_number, 'USD')   # FIXME - replace hardcoded 'USD'
            new_id = max([0] + [x['id'] for x in self._data[FOF.TRADES]]) + 1
            trade = {"id": new_id, "number": deal_number, "timestamp": timestamp, "settlement": settlement,
                     "account": account_id, "asset": asset_id, "quantity": qty, "price": price, "fee": fee}
            self._data[FOF.TRADES].append(trade)
            cnt += 1
            row += 1
        logging.info(self.tr("Stock trades loaded: ") + f"{cnt}")

    def _load_crypto_deals(self):
        cnt = 0
        columns = {
            "number": "Trade number \(Номер сделки\)",
            "timestamp": "Trade date \(Дата сделки\)",
            "settlement": "Settle date \(дата расчетов\)",
            "account_currency": "Account currency \(Валюта счета\)",
            "asset_name": "Description",
            "B/S": "Type of Transaction\(Тип сделки\)",
            "qty": "Quantity \(Кол-во\)",
            "fee": "Broker fees in account currency \(комиссия брокера в валюте счета\)",
            "amount": "Net Value in account currency \(Итого в валюте счета\)",
            "fee_ex": "Other fees in account currency \(прочие комиссии в валюте счета\)"
        }

        row, headers = self.find_section_start("Сделки c виртуальными \(крипто\) инструментами", columns,
                                               header_height=3)
        if row < 0:
            return
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':
                break
            # Dates are stored as date in Excel report file for crypto deals
            timestamp = int(self._statement[headers['timestamp']][row].replace(tzinfo=timezone.utc).timestamp())
            settlement = int(self._statement[headers['settlement']][row].replace(tzinfo=timezone.utc).timestamp())
            deal_number = str(self._statement[headers['number']][row])
            asset_id = self.asset_id({'type': FOF.ASSET_CRYPTO, 'symbol': self._statement[headers['asset_name']][row],
                                      'name': self._statement[headers['asset_name']][row],
                                      'currency': self.currency_id('USD')})  # FIXME - replace hardcoded 'USD'
            if self._statement[headers['B/S']][row].startswith('Купля'):
                qty = self._statement[headers['qty']][row]
            elif self._statement[headers['B/S']][row].startswith('Продажа'):
                qty = -self._statement[headers['qty']][row]
            else:
                row += 1
                logging.warning(self.tr("Unknown trade type: ") + self._statement[headers['B/S']][row])
                continue
            # Collect fees
            fee = self._statement[headers['fee']][row] if self._statement[headers['fee']][row] else 0.0
            if self._statement[headers['fee_ex']][row]:
                fee += self._statement[headers['fee_ex']][row]
            # Calculate price in account currency
            amount = self._statement[headers['amount']][row]
            price = -(amount + fee) / qty
            assert price > 0.0
            account_id = self._find_account_id(self._account_number, self._statement[headers['account_currency']][row])
            new_id = max([0] + [x['id'] for x in self._data[FOF.TRADES]]) + 1
            trade = {"id": new_id, "number": deal_number, "timestamp": timestamp, "settlement": settlement,
                     "account": account_id, "asset": asset_id, "quantity": qty, "price": price, "fee": fee}
            self._data[FOF.TRADES].append(trade)
            cnt += 1
            row += 1
        logging.info(self.tr("Crypto trades loaded: ") + f"{cnt}")

    def _load_cash_transactions(self):
        self._load_money_table()
        self._load_fees()

    def _load_money_table(self):
        columns = {
            "date": "Дата",
            "type": "Зачисление/списание",
            "amount": "Сумма",
            "description": "Описание",
            "note": "Комментарий"
        }
        operations = {
            '': None,
            'Переоценка': None,
            'Корпоративные действия::Дивиденды': self.dividend,
            'Внешние затраты::Комиссия внешнего брокера::Удержанный налог': None,  # Loaded in 2nd loop later
            'Внешние затраты::Комиссия внешнего депозитария': self.fee,
            'Перевод на ТП': self.transfer_in,
            'Списание c ТП': self.transfer_out,
            'Корпоративные действия::Компенсации': self.skip_warning,
            'Корректировка': self.skip_warning
        }

        start_row, headers = self.find_section_start("Движение денежных средств", columns)
        if start_row < 0:
            return
        cnt = 0
        row = start_row   # Process dividend rows
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':
                break
            try:
                timestamp = int(self._statement[headers['date']][row].replace(tzinfo=timezone.utc).timestamp())
            except TypeError:  # Skip 'Итого' and similar lines
                row += 1
                continue
            operation = self._statement[headers['description']][row]
            if operation not in operations:
                raise Statement_ImportError(self.tr("Unsuppported cash transaction ") + f"'{operation}'")
            account_id = self._find_account_id(self._account_number, 'USD')  # FIXME - replace hardcoded 'USD'
            if self._statement[headers['type']][row] == 'IN':
                amount = self._statement[headers['amount']][row]
            elif self._statement[headers['type']][row] == 'OUT':
                amount = -self._statement[headers['amount']][row]
            else:
                raise Statement_ImportError(self.tr("Unknown cash transaction type ") +
                                            f"'{self._statement[headers['type']][row]}'")
            if operations[operation] is not None:
                operations[operation](timestamp, account_id, amount, self._statement[headers['note']][row])
            cnt += 1
            row += 1

        row = start_row  # Process tax rows - should be done separately after all dividends
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':
                break
            if self._statement[headers['description']][row] != "Внешние затраты::Комиссия внешнего брокера::Удержанный налог":
                row += 1
                continue
            assert self._statement[headers['type']][row] == 'OUT'  # Should be outgoing flow
            timestamp = int(self._statement[headers['date']][row].replace(tzinfo=timezone.utc).timestamp())
            amount = self._statement[headers['amount']][row]
            self.tax(timestamp, amount, self._statement[headers['note']][row])
            cnt += 1
            row += 1
        logging.info(self.tr("Cash operations loaded: ") + f"{cnt}")

    # Locate asset by its full name either in loaded JSON data (first) or in JAL database (next)
    def _find_asset_by_name(self, asset_name) -> int:
        candidates = [x for x in self._data[FOF.ASSETS] if 'name' in x and x['name'] == asset_name]
        if len(candidates) == 1:
            return candidates[0]["id"]
        asset_id = JalAsset(data={'name': asset_name}, search=True, create=False).id()
        return -asset_id  # Negative value to indicate that asset was found in db

    # This method finds dividend with given parameters in already loaded JSON data
    def _locate_dividend(self, asset_id, timestamp, ex_date):
        for dividend in self._data[FOF.ASSET_PAYMENTS]:
            if dividend['type'] == FOF.PAYMENT_DIVIDEND and dividend['timestamp'] == timestamp and \
                    dividend['ex_date'] == ex_date and dividend['asset'] == asset_id:
                return dividend
        return None

    def _load_fees(self):
        cnt = 0
        columns = {
            "date": "Дата",
            "amount": "Сумма",
            "currency": "Валюта",
            "note": "Комментарий"
        }
        row, headers = self.find_section_start("Брокерская комиссия, удержанная за период", columns)
        if row < 0:
            return
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':
                break
            try:
                timestamp = int(self._statement[headers['date']][row].replace(tzinfo=timezone.utc).timestamp())
            except TypeError:
                break   # Stop processing if we encounter invalid date (supposed to be "Итого:" line)
            account_id = self._find_account_id(self._account_number, self._statement[headers['currency']][row])
            self.fee(timestamp, account_id, -self._statement[headers['amount']][row],
                     self._statement[headers['note']][row])
            cnt += 1
            row += 1

    def dividend(self, timestamp, account_id, amount, note):
        DividendPattern = r"Дивиденды;\s+(Начисление дивидендов полученных по счету.*\.\s+)?Инструмент\s+(?P<asset>.*);\s+Дата отсечки\s+(?P<date>.*)"
        parts = re.match(DividendPattern, note, re.IGNORECASE)  # FIXME - below code used in ibkr.py as well
        if parts is None:
            raise Statement_ImportError(self.tr("Can't parse Dividend description ") + f"'{note}'")
        dividend = parts.groupdict()
        if len(dividend) != DividendPattern.count("(?P<"):  # check that expected number of groups was matched
            raise Statement_ImportError(self.tr("Dividend description miss some data ") + f"'{note}'")
        asset_id = self._find_asset_by_name(dividend['asset'])
        ex_date = int(datetime.strptime(dividend['date'], "%d/%m/%Y").replace(tzinfo=timezone.utc).timestamp())
        new_id = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
        payment = {"id": new_id, "type": FOF.PAYMENT_DIVIDEND, "account": account_id, "timestamp": timestamp,
                   "ex_date": ex_date, "asset": asset_id, "amount": amount, "description": note}
        self._data[FOF.ASSET_PAYMENTS].append(payment)

    def tax(self, timestamp, amount, note):
        TaxPattern = r"Налог на дивиденды;\s+(Начисление дивидендов полученных по счету.*\.\s+)?Инструмент\s+(?P<asset>.*);\s+Дата отсечки\s+(?P<date>.*)"
        parts = re.match(TaxPattern, note, re.IGNORECASE)  # FIXME - below code used in ibkr.py as well
        if parts is None:
            raise Statement_ImportError(self.tr("Can't parse Dividend description ") + f"'{note}'")
        tax = parts.groupdict()
        if len(tax) != TaxPattern.count("(?P<"):  # check that expected number of groups was matched
            raise Statement_ImportError(self.tr("Dividend description miss some data ") + f"'{note}'")
        asset_id = self._find_asset_by_name(tax['asset'])
        ex_date = int(datetime.strptime(tax['date'], "%d/%m/%Y").replace(tzinfo=timezone.utc).timestamp())
        dividend_record = self._locate_dividend(asset_id, timestamp, ex_date)
        if dividend_record is None:
            raise Statement_ImportError(self.tr("Dividend for tax was not found ") + f"'{note}'")
        else:
            dividend_record['tax'] = amount

    def fee(self, timestamp, account_id, amount, note):
        new_id = max([0] + [x['id'] for x in self._data[FOF.INCOME_SPENDING]]) + 1
        fee = {"id": new_id, "timestamp": timestamp, "account": account_id, "peer": 0,
               "lines": [{"amount": amount, "category": -PredefinedCategory.Fees, "description": note}]}
        self._data[FOF.INCOME_SPENDING].append(fee)

    def transfer_in(self, timestamp, account_id, amount, note):
        account = [x for x in self._data[FOF.ACCOUNTS] if x["id"] == account_id][0]
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfer = {"id": new_id, "account": [0, account_id, 0],
                    "asset": [account['currency'], account['currency']], "timestamp": timestamp,
                    "withdrawal": amount, "deposit": amount, "fee": 0.0, "description": note}
        self._data[FOF.TRANSFERS].append(transfer)

    def transfer_out(self, timestamp, account_id, amount, note):
        account = [x for x in self._data[FOF.ACCOUNTS] if x["id"] == account_id][0]
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfer = {"id": new_id, "account": [account_id, 0, 0],
                    "asset": [account['currency'], account['currency']], "timestamp": timestamp,
                    "withdrawal": -amount, "deposit": -amount, "fee": 0.0, "description": note}
        self._data[FOF.TRANSFERS].append(transfer)

    def skip_warning(self, _timestamp, _account_id, amount, note):
        logging.warning(self.tr("Import skipped of transaction: ") + f"'{note}' ({amount})")



import os
import logging
import importlib

from PySide6.QtWidgets import QFileDialog
from PySide6.QtCore import QObject
from jal.constants import Setup
from jal.db.helpers import get_app_path
from jal.db.settings import JalSettings, FolderFor
from jal.data_export.xlsx import XLSX


class Reports(QObject):
    def __init__(self, parent, MdiArea):
        super().__init__()
        self.parent = parent
        self._mdi = MdiArea

        self.items = []
        self.loadReportsList()

    def mdi_area(self):
        return self._mdi

    def loadReportsList(self):
        reports_folder = get_app_path() + Setup.REPORT_PATH
        report_modules = [filename[:-3] for filename in os.listdir(reports_folder) if filename.endswith(".py")]
        for module_name in report_modules:
            logging.debug(f"Trying to load report module: {module_name}")
            module = importlib.import_module(f"jal.reports.{module_name}")
            try:
                report_class_name = getattr(module, "JAL_REPORT_CLASS")
            except AttributeError:
                continue
            try:
                class_instance = getattr(module, report_class_name)
            except AttributeError:
                logging.error(self.tr("Report class can't be loaded: ") + report_class_name)
                continue
            report = class_instance()
            group = report.group if hasattr(report, "group") else ''
            self.items.append({'group': group, 'name': report.name,
                               'module': module, 'window_class': report.window_class})
            logging.debug(f"Report class '{report_class_name}' providing '{report.name}' report has been loaded")
        self.items = sorted(self.items, key=lambda item: item['name'])

    # method is called directly from menu, so it contains QAction that was triggered
    def show(self, action):
        report_loader = self.items[action.data()]
        module = report_loader['module']
        class_instance = getattr(module, report_loader['window_class'])
        report = class_instance(self)
        self._mdi.addSubWindow(report, maximized=True)

    def show_report(self, window_class, settings, maximized=False):
        report = [x for x in self.items if x['window_class'] == window_class]
        if len(report) != 1:
            logging.warning(self.tr("Report not found for window class: ") + window_class)
            return
        report_loader = report[0]
        module = report_loader['module']
        class_instance = getattr(module, report_loader['window_class'])
        report = class_instance(self, settings)
        self._mdi.addSubWindow(report, maximized=maximized)

    # Save report content from the model to xls-file chosen by the user
    def save_report(self, name, model):
        folder = JalSettings().getRecentFolder(FolderFor.Report, '.')
        filename, filter = QFileDialog.getSaveFileName(self._mdi, self.tr("Save report to:"),
                                                       folder, self.tr("Excel files (*.xlsx)"))
        if filename:
            if filter == self.tr("Excel files (*.xlsx)") and filename[-5:] != '.xlsx':
                filename = filename + '.xlsx'
        else:
            return
        JalSettings().setRecentFolder(FolderFor.Report, filename)
        report = XLSX(filename)
        report.output_model(name, model)
        report.save()
        logging.info(self.tr("Report was saved to file ") + f"'{filename}'")



import logging
from datetime import datetime
from dateutil import tz

from PySide6.QtCore import Qt, Slot, QStringListModel, QByteArray
from PySide6.QtWidgets import QLabel, QDateTimeEdit, QLineEdit, QComboBox, QTableView, QHeaderView, QPushButton
from PySide6.QtSql import QSqlTableModel
from jal.widgets.abstract_operation_details import AbstractOperationDetails
from jal.widgets.reference_selector import AccountSelector, AssetSelector
from jal.widgets.delegates import WidgetMapperDelegateBase, AssetSelectorDelegate, FloatDelegate
from jal.db.view_model import JalViewModel
from jal.db.helpers import load_icon
from jal.db.operations import LedgerTransaction


# ----------------------------------------------------------------------------------------------------------------------
class CorporateActionWidgetDelegate(WidgetMapperDelegateBase):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.delegates = {'timestamp': self.timestamp_delegate,
                          'asset_id': self.symbol_delegate,
                          'qty': self.decimal_delegate}


# ----------------------------------------------------------------------------------------------------------------------
class CorporateActionWidget(AbstractOperationDetails):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.name = self.tr("Corporate Action")
        self.operation_type = LedgerTransaction.CorporateAction
        self.combo_model = None

        self.asset_delegate = AssetSelectorDelegate()
        self.float_delegate = FloatDelegate(2)
        self.percent_delegate = FloatDelegate(2, percent=True)

        self.date_label = QLabel(self)
        self.account_label = QLabel(self)
        self.type_label = QLabel(self)
        self.number_label = QLabel(self)
        self.before_label = QLabel(self)
        self.asset_label = QLabel(self)
        self.qty_label = QLabel(self)
        self.after_label = QLabel(self)
        self.comment_label = QLabel(self)
        self.arrow = QLabel(self)

        self.main_label.setText(self.name)
        self.date_label.setText(self.tr("Date/Time"))
        self.account_label.setText(self.tr("Account"))
        self.type_label.setText(self.tr("Type"))
        self.number_label.setText(self.tr("#"))
        self.asset_label.setText(self.tr("Asset"))
        self.qty_label.setText(self.tr("Qty"))
        self.comment_label.setText(self.tr("Note"))
        self.arrow.setText(" 🡆 ")

        self.timestamp_editor = QDateTimeEdit(self)
        self.timestamp_editor.setCalendarPopup(True)
        self.timestamp_editor.setTimeSpec(Qt.UTC)
        self.timestamp_editor.setFixedWidth(self.timestamp_editor.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.25)
        self.timestamp_editor.setDisplayFormat("dd/MM/yyyy hh:mm:ss")
        self.type = QComboBox(self)
        self.account_widget = AccountSelector(self)
        self.asset_widget = AssetSelector(self)
        self.qty_edit = QLineEdit(self)
        self.number = QLineEdit(self)
        self.comment = QLineEdit(self)
        self.add_button = QPushButton(load_icon("add.png"), '', self)
        self.add_button.setToolTip(self.tr("Add asset"))
        self.del_button = QPushButton(load_icon("remove.png"), '', self)
        self.del_button.setToolTip(self.tr("Remove asset"))
        self.results_table = QTableView(self)
        self.results_table.horizontalHeader().setFont(self.bold_font)
        self.results_table.setAlternatingRowColors(True)
        self.results_table.verticalHeader().setVisible(False)
        self.results_table.verticalHeader().setMinimumSectionSize(20)
        self.results_table.verticalHeader().setDefaultSectionSize(20)

        self.layout.addWidget(self.date_label, 1, 0, 1, 1, Qt.AlignLeft)
        self.layout.addWidget(self.type_label, 2, 0, 1, 1, Qt.AlignLeft)
        self.layout.addWidget(self.number_label, 3, 0, 1, 1, Qt.AlignRight)
        self.layout.addWidget(self.comment_label, 4, 0, 1, 6, Qt.AlignLeft)

        self.layout.addWidget(self.timestamp_editor, 1, 1, 1, 1)
        self.layout.addWidget(self.type, 2, 1, 1, 1)
        self.layout.addWidget(self.number, 3, 1, 1, 1)
        self.layout.addWidget(self.comment, 4, 1, 1, 3)

        self.layout.addWidget(self.account_label, 1, 2, 1, 1, Qt.AlignRight)
        self.layout.addWidget(self.asset_label, 2, 2, 1, 1, Qt.AlignRight)
        self.layout.addWidget(self.qty_label, 3, 2, 1, 1, Qt.AlignRight)

        self.layout.addWidget(self.account_widget, 1, 3, 1, 1)
        self.layout.addWidget(self.asset_widget, 2, 3, 1, 1)
        self.layout.addWidget(self.qty_edit, 3, 3, 1, 1)

        self.layout.addWidget(self.arrow, 2, 4, 2, 1)
        self.layout.addWidget(self.results_table, 2, 5, 3, 4)

        self.layout.addWidget(self.commit_button, 0, 7, 1, 1)
        self.layout.addWidget(self.revert_button, 0, 8, 1, 1)
        self.layout.addWidget(self.add_button, 1, 7, 1, 1)
        self.layout.addWidget(self.del_button, 1, 8, 1, 1)

        self.layout.addItem(self.verticalSpacer, 7, 5, 1, 1)
        self.layout.addItem(self.horizontalSpacer, 1, 6, 1, 1)

        self.add_button.clicked.connect(self.addResult)
        self.del_button.clicked.connect(self.delResult)

        super()._init_db("asset_actions")
        self.combo_model = QStringListModel([self.tr("N/A"),
                                             self.tr("Merger"),
                                             self.tr("Spin-Off"),
                                             self.tr("Symbol change"),
                                             self.tr("Split"),
                                             self.tr("Delisting")])
        self.type.setModel(self.combo_model)

        self.mapper.setItemDelegate(CorporateActionWidgetDelegate(self.mapper))

        self.results_model = ResultsModel(self.results_table, "action_results")
        self.results_model.setEditStrategy(QSqlTableModel.OnManualSubmit)
        self.results_table.setModel(self.results_model)
        self.results_model.dataChanged.connect(self.onDataChange)

        self.account_widget.changed.connect(self.mapper.submit)
        self.asset_widget.changed.connect(self.mapper.submit)

        self.mapper.addMapping(self.timestamp_editor, self.model.fieldIndex("timestamp"))
        self.mapper.addMapping(self.account_widget, self.model.fieldIndex("account_id"))
        self.mapper.addMapping(self.asset_widget, self.model.fieldIndex("asset_id"))
        self.mapper.addMapping(self.number, self.model.fieldIndex("number"))
        self.mapper.addMapping(self.qty_edit, self.model.fieldIndex("qty"))
        self.mapper.addMapping(self.comment, self.model.fieldIndex("note"))
        self.mapper.addMapping(self.type, self.model.fieldIndex("type"), QByteArray().setRawData("currentIndex", 12))

        self.results_table.setItemDelegateForColumn(2, self.asset_delegate)
        self.results_table.setItemDelegateForColumn(3, self.float_delegate)
        self.results_table.setItemDelegateForColumn(4, self.percent_delegate)

        self.model.select()
        self.results_model.select()
        self.results_model.configureView()

    def setId(self, id):
        super().setId(id)
        self.results_model.setFilter(f"action_results.action_id = {id}")

    @Slot()
    def addResult(self):
        new_record = self.results_model.record()
        new_record.setValue("qty", 0)
        new_record.setValue("value_share", 0)
        if not self.results_model.insertRecord(-1, new_record):
            logging.fatal(self.tr("Failed to add new record: ") + self.results_model.lastError().text())
            return

    @Slot()
    def delResult(self):
        selection = self.results_table.selectionModel().selection().indexes()
        for idx in selection:
            self.results_model.removeRow(idx.row())
            self.onDataChange(idx, idx, None)

    @Slot()
    def saveChanges(self):
        self.model.database().transaction()
        try:
            if not self.model.submitAll():
                raise RuntimeError(self.tr("Operation submit failed: ") + self.model.lastError().text())
            oid = self.model.data(self.model.index(0, self.model.fieldIndex("id")))
            if oid is None:  # we just have saved new action record and need last inserted id
                oid = self.model.last_insert_id()
            for row in range(self.results_model.rowCount()):
                self.results_model.setData(self.results_model.index(row, self.results_model.fieldIndex("action_id")), oid)
            if not self.results_model.submitAll():
                raise RuntimeError(self.tr("Operation details submit failed: ") + self.results_model.lastError().text())
        except Exception as e:
            self.model.database().rollback()
            logging.fatal(e)
            return
        self.modified = False
        self.commit_button.setEnabled(False)
        self.revert_button.setEnabled(False)
        self.dbUpdated.emit()

    @Slot()
    def revertChanges(self):
        self.model.revertAll()
        self.results_model.revertAll()
        self.modified = False
        self.commit_button.setEnabled(False)
        self.revert_button.setEnabled(False)

    def createNew(self, account_id=0):
        super().createNew(account_id)
        self.results_model.setFilter(f"action_results.action_id = 0")

    def prepareNew(self, account_id):
        new_record = super().prepareNew(account_id)
        new_record.setValue("timestamp", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        new_record.setValue("number", '')
        new_record.setValue("account_id", account_id)
        new_record.setValue("type", 0)
        new_record.setValue("asset_id", 0)
        new_record.setValue("qty", '0')
        new_record.setValue("note", None)
        return new_record

    def copyNew(self):
        super().copyNew()
        child_records = []
        for row in range(self.results_model.rowCount()):
            child_records.append(self.results_model.record(row))
        self.results_model.setFilter(f"action_results.action_id = 0")
        for record in reversed(child_records):
            record.setNull("id")
            record.setNull("action_id")
            assert self.results_model.insertRows(0, 1)
            self.results_model.setRecord(0, record)

    def copyToNew(self, row):
        new_record = self.model.record(row)
        new_record.setNull("id")
        new_record.setValue("timestamp", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        new_record.setValue("number", '')
        return new_record


# ----------------------------------------------------------------------------------------------------------------------
class ResultsModel(JalViewModel):
    def __init__(self, parent_view, table_name):
        super().__init__(parent_view, table_name)
        self._columns = ["id", "action_id", self.tr("Asset"), self.tr("Qty"), self.tr("Share, %")]

    def configureView(self):
        self._view.setColumnHidden(0, True)
        self._view.setColumnHidden(1, True)
        self._view.setColumnWidth(3, 100)
        self._view.setColumnWidth(4, 100)
        self._view.horizontalHeader().setSectionResizeMode(2, QHeaderView.Stretch)



from functools import partial
from datetime import datetime
from PySide6.QtCore import Qt, Slot, QObject, QAbstractItemModel, QModelIndex
from PySide6.QtGui import QAction, QBrush
from PySide6.QtWidgets import QMenu
from jal.reports.reports import Reports
from jal.db.asset import JalAsset
from jal.ui.reports.ui_income_spending_report import Ui_IncomeSpendingReportWidget
from jal.constants import CustomColor
from jal.db.helpers import load_icon
from jal.db.category import JalCategory
from jal.widgets.helpers import is_signal_connected, month_list, month_start_ts, month_end_ts, \
    week_list, week_start_ts, week_end_ts, str2int
from jal.widgets.delegates import GridLinesDelegate, FloatDelegate
from jal.widgets.mdi import MdiWidget

JAL_REPORT_CLASS = "IncomeSpendingReport"

MONTHLY = 12
WEEKLY = 53


# ----------------------------------------------------------------------------------------------------------------------
class ReportTreeItem(QObject):
    def __init__(self, begin, end, item_id, name, periods=MONTHLY, parent=None):
        super().__init__()
        self._parent = parent
        if parent is not None:
            self._periods = parent.periods   # Child should have the same periodicity as parent
        else:
            self._periods = periods
        self._id = item_id
        self.name = name
        self._begin = begin
        self._end = end
        self._y_s = int(datetime.utcfromtimestamp(begin).strftime('%Y'))
        self._y_e = int(datetime.utcfromtimestamp(end).strftime('%Y'))
        if self._periods == MONTHLY:
            self._p_s = str2int(datetime.utcfromtimestamp(begin).strftime('%m'))
            self._p_e = str2int(datetime.utcfromtimestamp(end).strftime('%m'))
        elif self._periods == WEEKLY:
            self._p_s = str2int(datetime.utcfromtimestamp(begin).strftime('%W'))
            if self._p_s == 0:  # This is a week that starts a year before
                self._y_s -= 1
                self._p_s = 53
            self._p_e = str2int(datetime.utcfromtimestamp(end).strftime('%W'))
        else:
            assert False, "Wrong period for Income/Spending report"
        # amounts is 2D-array of per month amounts:
        # amounts[year][period_id] - amount for particular month (period_id from 1 to 12) or week (from 1 to 53)
        # amounts[year][0] - total per year
        self._amounts = [[0] * (self._periods + 1) for _ in range(self._y_e - self._y_s + 1)]
        self._total = 0
        self._children = []

    def appendChild(self, child):
        child.setParent(self)
        self._children.append(child)

    def getChild(self, id):
        if id < 0 or id > len(self._children):
            return None
        return self._children[id]

    def childrenCount(self):
        return len(self._children)

    def removeEmptyChildren(self):
        for child in self._children:
            child.removeEmptyChildren()
        self._children = [x for x in self._children if x.getAmount(0, 0) != 0]

    def dataCount(self):
        if self._y_s == self._y_e:
            return self._p_e - self._p_s + 3  # + 1 for year, + 1 for totals
        else:
            # (self._periods + 1) * (self._y_e - self._y_s - 1) + (self._p_e + 1) + (self._periods - self._p_s + 2) + 1 simplified to:
            return (self._periods + 1) * (self._y_e - self._y_s - 1) + (self._p_e - self._p_s + self._periods + 4)

    # Converts table column number into (year, period_id) value
    # column 0 - name of row - return (-1, -1)
    # then repetition of [year], [jan], [feb] ... [nov], [dec] - return (year, 0), (year, 1) ... (year, 12)
    # or repetition of [year], [wk1], [wk2] .. [wk53] - return (year, 0), (year, 1) ... (year, 53)
    # last column is total value - return (0, 0)
    def column2calendar(self, column: int) -> (int, int):
        if column == 0:
            return -1, -1
        if column == self.dataCount():
            return 0, 0
        if column == 1:
            return self._y_s, 0
        column = column + self._p_s - 2
        year = self._y_s + int(column / (self._periods + 1))
        period_id = column % (self._periods + 1)
        return year, period_id

    def period_name(self, period_id: int) -> str:
        month_name = [
            self.tr('Jan'), self.tr('Feb'), self.tr('Mar'), self.tr('Apr'), self.tr('May'), self.tr('Jun'),
            self.tr('Jul'), self.tr('Aug'), self.tr('Sep'), self.tr('Oct'), self.tr('Nov'), self.tr('Dec')
        ]
        if self._periods == MONTHLY:
            return month_name[period_id - 1]
        elif self._periods == WEEKLY:
            prefix = self.tr("wk")   # On separate line as it won't be extracted for translation otherwise
            return f"{prefix}{period_id}"
        else:
            assert False, "Wrong period for Income/Spending report"

    def setParent(self, parent):
        self._parent = parent

    def getParent(self):
        return self._parent

    def addAmount(self, year, period_id, amount):
        y_i = year - self._y_s
        self._amounts[y_i][period_id] += amount
        self._amounts[y_i][0] += amount
        self._total += amount
        if self._parent is not None:
            self._parent.addAmount(year, period_id, amount)

    # Return amount for given year and period or total amount if year==0
    def getAmount(self, year, period_id):
        if year == 0:
            return self._total
        y_i = year - self._y_s
        return self._amounts[y_i][period_id]

    def getLeafById(self, id):
        if self._id == id:
            return self
        leaf = None
        for child in self._children:
            leaf = child.getLeafById(id)
        return leaf

    def details(self, year, period_id):
        if period_id == 0:
            p_begin = 1
            p_end = self._periods
        else:
            p_begin = p_end = period_id
        if year == 0:
            begin_ts = self._begin
            end_ts = self._end
        elif self._periods == MONTHLY:
            begin_ts = month_start_ts(year, p_begin) if month_start_ts(year, p_begin) > self._begin else self._begin
            end_ts = month_end_ts(year, p_end) if month_end_ts(year, p_end) < self._end else self._end
        elif self._periods == WEEKLY:
            begin_ts = week_start_ts(year, p_begin) if week_start_ts(year, p_begin) > self._begin else self._begin
            end_ts = week_end_ts(year, p_end) if week_end_ts(year, p_end) < self._end else self._end
        else:
            assert False, "Wrong period for Income/Spending report"
        item_summary = {
            'category_id': self._id,
            'begin_ts': begin_ts,
            'end_ts': end_ts,
            'total': self.getAmount(year, period_id)
        }
        return item_summary

    @property
    def year_begin(self):
        return self._y_s

    @property
    def period_begin(self):
        return self._p_s

    @property
    def year_end(self):
        return self._y_e

    @property
    def period_end(self):
        return self._p_e

    @property
    def periods(self):
        return self._periods


# ----------------------------------------------------------------------------------------------------------------------
class IncomeSpendingReportModel(QAbstractItemModel):
    def __init__(self, parent_view):
        super().__init__(parent_view)
        self._periodicity = MONTHLY
        self._begin = 0
        self._end = 0
        self._currency = 0
        self._period_list = []
        self._view = parent_view
        self._root = None
        self._grid_delegate = None
        self._float_delegate = None

    def rowCount(self, parent=None):
        if not parent.isValid():
            parent_item = self._root
        else:
            parent_item = parent.internalPointer()
        if parent_item is not None:
            return parent_item.childrenCount()
        else:
            return 0

    def columnCount(self, parent=None):
        if parent is None:
            parent_item = self._root
        elif not parent.isValid():
            parent_item = self._root
        else:
            parent_item = parent.internalPointer()
        if parent_item is not None:
            return parent_item.dataCount() + 1  # + 1 for row header
        else:
            return 0

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal:
            if role == Qt.DisplayRole:
                year, period = self._root.column2calendar(section)
                if year < 0:
                    col_name = ''
                elif year == 0:
                    col_name = self.tr("TOTAL")
                else:
                    if period == 0:
                        status = '▼' if self._view.isColumnHidden(section + 1) else '▶'
                        col_name = f"{year} " + status
                    else:
                        col_name = self._root.period_name(period)
                return col_name
            if role == Qt.TextAlignmentRole:
                return int(Qt.AlignCenter)
        return None

    def headerWidth(self, section):
        return self._view.header().sectionSize(section)

    def index(self, row, column, parent=None):
        if not parent.isValid():
            parent = self._root
        else:
            parent = parent.internalPointer()
        child = parent.getChild(row)
        if child:
            return self.createIndex(row, column, child)
        return QModelIndex()

    def parent(self, index=None):
        if not index.isValid():
            return QModelIndex()
        child_item = index.internalPointer()
        parent_item = child_item.getParent()
        if parent_item == self._root:
            return QModelIndex()
        return self.createIndex(0, 0, parent_item)

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return None
        item = index.internalPointer()
        if role == Qt.DisplayRole:
            if index.column() == 0:
                return item.name
            else:
                year, period = self._root.column2calendar(index.column())
                return item.getAmount(year, period)
        if role == Qt.ForegroundRole:
            if index.column() != 0:
                year, period = self._root.column2calendar(index.column())
                if item.getAmount(year, period) > 0:
                    return QBrush(CustomColor.DarkGreen)
                elif item.getAmount(year, period) < 0:
                    return QBrush(CustomColor.DarkRed)
                else:
                    return QBrush(CustomColor.Grey)
        if role == Qt.TextAlignmentRole:
            if index.column() == 0:
                return int(Qt.AlignLeft)
            else:
                return int(Qt.AlignRight)
        if role == Qt.UserRole:  # return category id for given index
            if index.column() != 0:
                year, period = self._root.column2calendar(index.column())
                return item.details(year, period)
        return None

    def configureView(self):
        self._grid_delegate = GridLinesDelegate(self._view)
        self._float_delegate = FloatDelegate(2, allow_tail=False, parent=self._view)
        for column in range(self.columnCount()):
            if column == 0:
                self._view.setItemDelegateForColumn(column, self._grid_delegate)
                self._view.setColumnWidth(column, 300)
            else:
                self._view.setItemDelegateForColumn(column, self._float_delegate)
                self._view.setColumnWidth(column, 100)
        font = self._view.header().font()
        font.setBold(True)
        self._view.header().setFont(font)

        if not is_signal_connected(self._view.header(), "sectionDoubleClicked(int)"):
            self._view.header().sectionDoubleClicked.connect(self.toggeYearColumns)

    def toggeYearColumns(self, section):
        year, period = self._root.column2calendar(section)
        if year >= 0 and period == 0:
            if year == self._root.year_begin:
                year_columns = 12 - self._root.period_begin + 1
            elif year == self._root.year_end:
                year_columns = self._root.period_end
            else:
                year_columns = 12
            for i in range(year_columns):
                new_state = not self._view.isColumnHidden(section + i + 1)
                self._view.setColumnHidden(section + i + 1, new_state)
            self.headerDataChanged.emit(Qt.Horizontal, section, section)

    def setPeriod(self, period_id):
        if period_id == 1:
            self._periodicity = WEEKLY
        else:
            self._periodicity = MONTHLY
        self.prepareData()
        self.configureView()

    def setDatesRange(self, begin, end):
        self._begin = begin
        self._end = end
        self.prepareData()
        self.configureView()

    def setCurrency(self, currency_id):
        if self._currency != currency_id:
            self._currency = currency_id
            self.prepareData()
            self.configureView()

    def prepareData(self):
        if not self._currency:
            return
        root_category = JalCategory(0)
        if self._periodicity == MONTHLY:
            self._period_list = month_list(self._begin, self._end)
        elif self._periodicity == WEEKLY:
            self._period_list = week_list(self._begin, self._end)
        else:
            assert False, "Wrong period for Income/Spending report"
        self._root = ReportTreeItem(self._begin, self._end, -1, "ROOT", periods=self._periodicity)  # invisible root
        self._root.appendChild(ReportTreeItem(self._begin, self._end, 0, self.tr("TOTAL"), periods=self._periodicity))  # visible root
        self._load_child_amounts(root_category)
        self._root.removeEmptyChildren()
        self.modelReset.emit()
        self._view.expandAll()

    def _load_child_amounts(self, parent_category: JalCategory):
        for category in parent_category.get_child_categories():
            leaf = self._root.getLeafById(category.id())
            if leaf is None:
                parent = self._root.getLeafById(category.parent_id())
                leaf = ReportTreeItem(self._begin, self._end, category.id(), category.name(), parent=parent)
                parent.appendChild(leaf)
            for period in self._period_list:
                leaf.addAmount(period['year'], period['number'],
                               category.get_turnover(period['begin_ts'], period['end_ts'], self._currency))
            self._load_child_amounts(category)


# ----------------------------------------------------------------------------------------------------------------------
class IncomeSpendingReport(QObject):
    def __init__(self):
        super().__init__()
        self.name = self.tr("Income & Spending")
        self.window_class = "IncomeSpendingReportWindow"


# ----------------------------------------------------------------------------------------------------------------------
class IncomeSpendingReportWindow(MdiWidget):
    def __init__(self, parent: Reports, settings: dict = None):
        super().__init__(parent.mdi_area())
        self.ui = Ui_IncomeSpendingReportWidget()
        self.ui.setupUi(self)
        self._parent = parent
        self.name = self.tr("Income & Spending")
        self.current_index = None  # this is used in onOperationContextMenu() to track item for menu

        self.income_spending_model = IncomeSpendingReportModel(self.ui.ReportTreeView)
        self.ui.ReportTreeView.setModel(self.income_spending_model)
        self.ui.ReportTreeView.setContextMenuPolicy(Qt.CustomContextMenu)

        # Operations view context menu
        self.contextMenu = QMenu(self.ui.ReportTreeView)
        self.actionDetails = QAction(load_icon("list.png"), self.tr("Show operations"), self)
        self.contextMenu.addAction(self.actionDetails)

        self.connect_signals_and_slots()

        if settings is None:
            begin, end = self.ui.ReportRange.getRange()
            period = self.ui.PeriodComboBox.currentIndex()
            settings = {'period': period, 'begin_ts': begin, 'end_ts': end, 'currency_id': JalAsset.get_base_currency()}
        self.ui.PeriodComboBox.setCurrentIndex(settings['period'])
        self.ui.ReportRange.setRange(settings['begin_ts'], settings['end_ts'])
        self.ui.CurrencyCombo.setIndex(settings['currency_id'])

    def connect_signals_and_slots(self):
        self.ui.PeriodComboBox.currentIndexChanged.connect(self.ui.ReportTreeView.model().setPeriod)
        self.ui.ReportRange.changed.connect(self.ui.ReportTreeView.model().setDatesRange)
        self.ui.CurrencyCombo.changed.connect(self.ui.ReportTreeView.model().setCurrency)
        self.ui.ReportTreeView.customContextMenuRequested.connect(self.onCellContextMenu)
        self.actionDetails.triggered.connect(self.showDetailsReport)
        self.ui.SaveButton.pressed.connect(partial(self._parent.save_report, self.name, self.ui.ReportTreeView.model()))

    @Slot()
    def onCellContextMenu(self, position):
        self.current_index = self.ui.ReportTreeView.indexAt(position)
        if self.current_index.isValid() and self.current_index.column() != 0:
            self.contextMenu.popup(self.ui.ReportTreeView.viewport().mapToGlobal(position))

    @Slot()
    def showDetailsReport(self):
        details = self.income_spending_model.data(self.current_index, Qt.UserRole)
        self._parent.show_report("CategoryReportWindow", details)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'select_reference_dlg.ui'
##
## Created by: Qt User Interface Compiler version 6.5.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QDialog, QDialogButtonBox,
    QFrame, QLabel, QSizePolicy, QSpacerItem,
    QVBoxLayout, QWidget)

class Ui_SelectReferenceDlg(object):
    def setupUi(self, SelectReferenceDlg):
        if not SelectReferenceDlg.objectName():
            SelectReferenceDlg.setObjectName(u"SelectReferenceDlg")
        SelectReferenceDlg.resize(400, 97)
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(SelectReferenceDlg.sizePolicy().hasHeightForWidth())
        SelectReferenceDlg.setSizePolicy(sizePolicy)
        SelectReferenceDlg.setWindowTitle(u"Title")
        self.WindowLayout = QVBoxLayout(SelectReferenceDlg)
        self.WindowLayout.setObjectName(u"WindowLayout")
        self.DescriptionLabel = QLabel(SelectReferenceDlg)
        self.DescriptionLabel.setObjectName(u"DescriptionLabel")
        self.DescriptionLabel.setText(u"Description")

        self.WindowLayout.addWidget(self.DescriptionLabel)

        self.SelectorFrame = QFrame(SelectReferenceDlg)
        self.SelectorFrame.setObjectName(u"SelectorFrame")
        self.SelectorFrame.setFrameShape(QFrame.NoFrame)
        self.SelectorFrame.setFrameShadow(QFrame.Plain)
        self.SelectorFrame.setLineWidth(0)
        self.SelectorFrame.setMidLineWidth(0)
        self.FrameLayout = QVBoxLayout(self.SelectorFrame)
        self.FrameLayout.setSpacing(0)
        self.FrameLayout.setObjectName(u"FrameLayout")
        self.FrameLayout.setContentsMargins(0, 0, 0, 0)

        self.WindowLayout.addWidget(self.SelectorFrame)

        self.buttonBox = QDialogButtonBox(SelectReferenceDlg)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.WindowLayout.addWidget(self.buttonBox)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.WindowLayout.addItem(self.verticalSpacer)


        self.retranslateUi(SelectReferenceDlg)
        self.buttonBox.accepted.connect(SelectReferenceDlg.close)
        self.buttonBox.rejected.connect(SelectReferenceDlg.reject)

        QMetaObject.connectSlotsByName(SelectReferenceDlg)
    # setupUi

    def retranslateUi(self, SelectReferenceDlg):
        pass
    # retranslateUi




import decimal
from functools import partial
from decimal import Decimal

from PySide6.QtCore import Qt, Slot, QObject, QAbstractTableModel
from jal.ui.reports.ui_profit_loss_report import Ui_ProfitLossReportWidget
from jal.reports.reports import Reports
from jal.db.account import JalAccount
from jal.db.asset import JalAsset
from jal.constants import BookAccount, PredefinedCategory
from jal.widgets.helpers import month_list
from jal.widgets.delegates import FloatDelegate
from jal.widgets.mdi import MdiWidget

JAL_REPORT_CLASS = "ProfitLossReport"


#-----------------------------------------------------------------------------------------------------------------------
class ProfitLossModel(QAbstractTableModel):
    def __init__(self, parent_view):
        super().__init__(parent_view)
        self._columns = [self.tr("Period"), self.tr("Money"), self.tr("In / Out"), self.tr("Dividends"), self.tr("%"),
                         self.tr("Fees"), self.tr("Taxes"), self.tr("Assets"), self.tr("P&L"), self.tr("Total"),
                         self.tr("Change"), self.tr("Change, %")]
        self.month_name = [
            self.tr('Jan'), self.tr('Feb'), self.tr('Mar'), self.tr('Apr'), self.tr('May'), self.tr('Jun'),
            self.tr('Jul'), self.tr('Aug'), self.tr('Sep'), self.tr('Oct'), self.tr('Nov'), self.tr('Dec')
        ]
        self._view = parent_view
        self._data = []
        self._month_list = []
        self._begin = 0
        self._end = 0
        self._account_id = 0
        self._float_delegate = None
        self._color_delegate = None
        self._percent_delegate = None

    def rowCount(self, parent=None):
        return len(self._data)

    def columnCount(self, parent=None):
        return len(self._columns)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._columns[section]
        return None

    def headerWidth(self, section):
        return self._view.horizontalHeader().sectionSize(section)

    def data(self, index, role=Qt.DisplayRole, field=''):
        if role == Qt.DisplayRole:
            return self._data[index.row()][index.column()]

    def setDatesRange(self, begin, end):
        self._begin = begin
        self._end = end
        self._month_list = month_list(begin, end)
        self.prepareData()
        self.configureView()

    def setAccount(self, account_id):
        self._account_id = account_id
        self.prepareData()
        self.configureView()

    # returns a dictionary with following keys (period is given by begin and end timestamps):
    # money - amount of money by end of the period
    # transfers - amount of money that came in(+) and out(-) of the account
    # dividends, interests, fees, taxes - amount of money that changed account value due to such events
    # assets - valuation of assets held by account by prices at the end of the period
    # p&l - profit and loss of deals closed during the period
    # total - total amount of money and assets by end of the period
    def data4period(self, begin: int, end: int, account: JalAccount) -> dict:
        assets = account.assets_list(end)
        asset_value = Decimal('0')
        for asset_data in assets:
            asset = asset_data['asset']
            asset_value += asset_data['amount'] * asset.quote(end, account.currency())[1]
        data = {
            'money': account.get_asset_amount(end, account.currency()),
            'transfers': -account.get_book_turnover(BookAccount.Transfers, begin, end),
            'dividends': -account.get_category_turnover(PredefinedCategory.Dividends, begin, end),
            'interest': -account.get_category_turnover(PredefinedCategory.Interest, begin, end),
            'fees': -account.get_category_turnover(PredefinedCategory.Fees, begin, end),
            'taxes': -account.get_category_turnover(PredefinedCategory.Taxes, begin, end),
            'assets': asset_value,
            'p&l': -account.get_category_turnover(PredefinedCategory.Profit, begin, end),
            'total': account.get_asset_amount(end, account.currency()) + asset_value
        }
        return data

    def prepareData(self):
        self._data = []
        money_p = assets_p = money_0 = assets_0 = None
        if not self._month_list:
            self.modelReset.emit()
            return
        account = JalAccount(self._account_id)
        # Prepend table with initial row and extend it with totals row
        months = [{'begin_ts': self._month_list[0]['begin_ts'], 'end_ts': self._month_list[0]['begin_ts']}]
        months.extend(self._month_list)
        months.append({'begin_ts': self._month_list[0]['begin_ts'], 'end_ts': self._month_list[-1]['end_ts']})
        for i, month in enumerate(months):
            values = self.data4period(month['begin_ts'], month['end_ts'], account)
            if i == 0:
                row_name = self.tr("Period start")
                money_p = money_0 = values['money']
                assets_p = assets_0 = values['assets']
            elif i == len(months) - 1:
                row_name = self.tr("Period end")
                money_p = money_0
                assets_p = assets_0
            else:
                row_name = f"{month['year']} {self.month_name[month['number'] - 1]}"
            try:
                rel_change = ((values['money'] + values['assets']) - (money_p + assets_p)) / (money_p + assets_p)
            except (ZeroDivisionError, decimal.InvalidOperation):
                rel_change = Decimal('0')
            data_row = [row_name, values['money'], values['transfers'], values['dividends'], values['interest'],
                        values['fees'], values['taxes'], values['assets'], values['p&l'],
                        values['total'], (values['money'] + values['assets']) - (money_p + assets_p),
                        rel_change]
            self._data.append(data_row)
            money_p = values['money']
            assets_p = values['assets']
        self.modelReset.emit()

    def configureView(self):
        self._view.setModel(self)
        font = self._view.horizontalHeader().font()
        font.setBold(True)
        self._view.horizontalHeader().setFont(font)
        self._float_delegate = FloatDelegate(2, allow_tail=False)
        self._color_delegate = FloatDelegate(2, allow_tail=False, colors=True)
        self._percent_delegate = FloatDelegate(2, allow_tail=False, colors=True, percent=True)
        self._view.setItemDelegateForColumn(1, self._float_delegate)
        self._view.setItemDelegateForColumn(2, self._color_delegate)
        self._view.setItemDelegateForColumn(3, self._float_delegate)
        self._view.setItemDelegateForColumn(4, self._float_delegate)
        self._view.setItemDelegateForColumn(5, self._float_delegate)
        self._view.setItemDelegateForColumn(6, self._float_delegate)
        self._view.setItemDelegateForColumn(7, self._float_delegate)
        self._view.setItemDelegateForColumn(8, self._color_delegate)
        self._view.setItemDelegateForColumn(9, self._float_delegate)
        self._view.setItemDelegateForColumn(10, self._color_delegate)
        self._view.setItemDelegateForColumn(11, self._percent_delegate)


# ----------------------------------------------------------------------------------------------------------------------
class ProfitLossReport(QObject):
    def __init__(self):
        super().__init__()
        self.name = self.tr("P&L by Account")
        self.window_class = "ProfitLossReportWindow"


# ----------------------------------------------------------------------------------------------------------------------
class ProfitLossReportWindow(MdiWidget):
    def __init__(self, parent: Reports, settings: dict = None):
        super().__init__(parent.mdi_area())
        self.ui = Ui_ProfitLossReportWidget()
        self.ui.setupUi(self)
        self._parent = parent
        self.name = self.tr("P&L by Account")

        self.pl_model = ProfitLossModel(self.ui.ReportTableView)
        self.ui.ReportTableView.setModel(self.pl_model)

        self.connect_signals_and_slots()

    def connect_signals_and_slots(self):
        self.ui.ReportRange.changed.connect(self.ui.ReportTableView.model().setDatesRange)
        self.ui.ReportAccountEdit.changed.connect(self.onAccountChange)
        self.ui.SaveButton.pressed.connect(partial(self._parent.save_report, self.name, self.ui.ReportTableView.model()))

    @Slot()
    def onAccountChange(self):
        account_id = self.ui.ReportAccountEdit.selected_id
        self.ui.ReportTableView.model().setAccount(account_id)
        self.ui.CurrencyLbl.setText(self.tr("Currency: ") + JalAsset(JalAccount(account_id).currency()).symbol())



from jal.db.db import JalDB
from PySide6.QtCore import QStandardPaths, QFileInfo


class FolderFor:
    Statement = 1
    Report = 2

class JalSettings(JalDB):
    __RECENT_PREFIX = "RecentFolder_"
    __folders = {
        FolderFor.Statement: "Statement",
        FolderFor.Report: "Report"
    }

    def __init__(self):
        super().__init__()

    def DbPath(self):
        return self._db_path()

    def getValue(self, key, default=None):
        value = self._read("SELECT value FROM settings WHERE name=:key", [(":key", key)])
        if value is None:
            value = default
        return value

    def setValue(self, key, value):
        self._exec("INSERT OR REPLACE INTO settings(id, name, value) "
                   "VALUES((SELECT id FROM settings WHERE name=:key), :key, :value)",
                   [(":key", key), (":value", value)], commit=True)

    # Returns 2-letter language code that corresponds to current 'Language' settings in DB
    def getLanguage(self):
        lang_id = self.getValue('Language', default=1)
        return self._read("SELECT language FROM languages WHERE id = :language_id", [(':language_id', lang_id)])

    # Set 'Language' setting in DB that corresponds to given 2-letter language code
    def setLanguage(self, language_code):
        lang_id = self._read("SELECT id FROM languages WHERE language = :language_code",
                             [(':language_code', language_code)])
        self.setValue('Language', lang_id)

    def getRecentFolder(self, folder_type: int, default: str=''):
        folder = self.getValue(self.__RECENT_PREFIX + self.__folders[folder_type])
        if not folder: folder = QStandardPaths.writableLocation(QStandardPaths.DocumentsLocation)
        if not folder: folder = default
        return folder

    def setRecentFolder(self, folder_type: int, folder: str):
        file_info = QFileInfo(folder)
        if file_info.isDir():
            path = file_info.absoluteFilePath()
        else:
            path = file_info.absolutePath()
        self.setValue(self.__RECENT_PREFIX + self.__folders[folder_type], path)



import logging
from datetime import datetime
from decimal import Decimal
from PySide6.QtCore import Qt, Property, QDateTime, QTimeZone, QLocale
from PySide6.QtSql import QSqlRelation, QSqlRelationalDelegate
from PySide6.QtWidgets import QDialog, QDataWidgetMapper, QStyledItemDelegate, QComboBox, QLineEdit
from jal.ui.ui_asset_dlg import Ui_AssetDialog
from jal.constants import PredefinedAsset, AssetData
from jal.db.helpers import load_icon, localize_decimal
from jal.widgets.delegates import DateTimeEditWithReset, BoolDelegate
from jal.db.reference_models import AbstractReferenceListModel
from jal.db.tag import JalTag
from jal.widgets.reference_selector import TagSelector


class AssetsListModel(AbstractReferenceListModel):
    def __init__(self, table, parent_view):
        super().__init__(table=table, parent_view=parent_view)
        self._columns = [("id", ''),
                         ("type_id", 'Asset type'),
                         ("full_name", self.tr("Asset name")),
                         ("isin", self.tr("ISIN")),
                         ("country_id", self.tr("Country")),
                         ("base_asset", self.tr("Base asset"))]


class AssetDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.ui = Ui_AssetDialog()
        self.ui.setupUi(self)
        self._asset_id = -1
        # Custom model to allow common submit errors handling and error message display
        self._model = AssetsListModel("assets", self)

        self._mapper = QDataWidgetMapper(self._model)
        self._mapper.setModel(self._model)
        self._mapper.setSubmitPolicy(QDataWidgetMapper.AutoSubmit)

        self._mapper.addMapping(self.ui.NameEdit, self._model.fieldIndex("full_name"))
        self._mapper.addMapping(self.ui.isinEdit, self._model.fieldIndex("isin"))
        self._mapper.addMapping(self.ui.TypeCombo, self._model.fieldIndex("type_id"))
        self._mapper.addMapping(self.ui.CountryCombo, self._model.fieldIndex("country_id"))
        self._mapper.addMapping(self.ui.BaseAssetSelector, self._model.fieldIndex("base_asset"))

        self._model.select()

        self._symbols_model = SymbolsListModel("asset_tickers", self.ui.SymbolsTable)
        self.ui.SymbolsTable.setModel(self._symbols_model)
        self._symbols_model.select()
        self._symbols_model.configureView()

        self._data_model = ExtraDataModel("asset_data", self.ui.DataTable)
        self.ui.DataTable.setModel(self._data_model)
        self._data_model.select()
        self._data_model.configureView()

        self.ui.AddSymbolButton.setIcon(load_icon("add.png"))
        self.ui.RemoveSymbolButton.setIcon(load_icon("delete.png"))
        self.ui.AddDataButton.setIcon(load_icon("add.png"))
        self.ui.RemoveDataButton.setIcon(load_icon("delete.png"))
        self.ui.OkButton.setIcon(load_icon("accept.png"))
        self.ui.CancelButton.setIcon(load_icon("cancel.png"))

        self.ui.TypeCombo.currentIndexChanged.connect(self.onTypeUpdate)
        self.ui.AddSymbolButton.clicked.connect(self.onAddSymbol)
        self.ui.RemoveSymbolButton.clicked.connect(self.onRemoveSymbol)
        self.ui.AddDataButton.clicked.connect(self.onAddData)
        self.ui.RemoveDataButton.clicked.connect(self.onRemoveData)

    def getSelectedId(self):
        return self._asset_id

    def setSelectedId(self, asset_id):
        self._asset_id = asset_id
        self._model.setFilter(f"id={self._asset_id}")
        self._mapper.toFirst()
        self._symbols_model.filterBy("asset_id", self._asset_id)
        self._data_model.filterBy("asset_id", self._asset_id)
        self.onTypeUpdate(0)   # need to update manually as it isn't triggered from mapper

    selected_id = Property(str, getSelectedId, setSelectedId)

    def createNewRecord(self):
        self._asset_id = 0
        self._model.setFilter(f"id={self._asset_id}")
        new_record = self._model.record()
        new_record.setNull("id")
        assert self._model.insertRows(0, 1)
        self._model.setRecord(0, new_record)
        self._mapper.toLast()
        self._symbols_model.filterBy("asset_id", self._asset_id)
        self._data_model.filterBy("asset_id", self._asset_id)

    def accept(self) -> None:
        self._model.database().transaction()
        try:
            if not self._model.submitAll():
                raise RuntimeError(self.tr("Asset submit failed: ") + self._model.lastError().text())
            asset_id = self._model.data(self._model.index(0, self._model.fieldIndex("id")))
            if asset_id is None:  # we just have saved new asset record and need last inserted id
                asset_id = self._model.last_insert_id()
            for model in [self._data_model, self._symbols_model]:
                for row in range(model.rowCount()):
                    model.setData(model.index(row, model.fieldIndex("asset_id")), asset_id)
                if not model.submitAll():
                    raise RuntimeError(self.tr("Asset details submit failed: ") + model.lastError().text())
        except Exception as e:
            self._model.database().rollback()
            logging.fatal(e)
            return
        self._asset_id = asset_id
        super().accept()

    def reject(self) -> None:
        for model in [self._data_model, self._symbols_model, self._model]:
            model.revertAll()
        super().reject()

    def onTypeUpdate(self, _index):
        if self.ui.TypeCombo.key == PredefinedAsset.Derivative:
            self.ui.BaseAssetSelector.setEnabled(True)
            self.ui.isinEdit.setEnabled(False)
        elif self.ui.TypeCombo.key == PredefinedAsset.Money or self.ui.TypeCombo.key == PredefinedAsset.Commodity:
            self.ui.BaseAssetSelector.setEnabled(False)
            self.ui.isinEdit.setEnabled(False)
        else:
            self.ui.BaseAssetSelector.setEnabled(False)
            self.ui.isinEdit.setEnabled(True)

    def onAddSymbol(self):
        idx = self.ui.SymbolsTable.selectionModel().selection().indexes()
        current_index = idx[0] if idx else self._symbols_model.index(0, 0)
        self._symbols_model.addElement(current_index)

    def onRemoveSymbol(self):
        idx = self.ui.SymbolsTable.selectionModel().selection().indexes()
        current_index = idx[0] if idx else self._symbols_model.index(0, 0)
        self._symbols_model.removeElement(current_index)

    def onAddData(self):
        idx = self.ui.DataTable.selectionModel().selection().indexes()
        current_index = idx[0] if idx else self._data_model.index(0, 0)
        self._data_model.addElement(current_index)
    def onRemoveData(self):
        idx = self.ui.DataTable.selectionModel().selection().indexes()
        current_index = idx[0] if idx else self._data_model.index(0, 0)
        self._data_model.removeElement(current_index)


class SymbolsListModel(AbstractReferenceListModel):
    def __init__(self, table, parent_view):
        super().__init__(table=table, parent_view=parent_view)
        self._columns = [("id", ''),
                         ("asset_id", ''),
                         ("symbol", self.tr("Symbol")),
                         ("currency_id", self.tr("Currency")),
                         ("description", self.tr("Description")),
                         ("quote_source", self.tr("Quotes")),
                         ("active", self.tr("Act."))]
        self._default_name = "symbol"
        self._sort_by = "symbol"
        self._hidden = ["id", "asset_id"]
        self._stretch = "description"
        self._lookup_delegate = None
        self._bool_delegate = None
        self._default_values = {'description': '', 'currency_id': 1, 'quote_source': -1, 'active': 1}
        self.setRelation(self.fieldIndex("currency_id"), QSqlRelation("currencies", "id", "symbol"))
        self.setRelation(self.fieldIndex("quote_source"), QSqlRelation("data_sources", "id", "name"))

    def configureView(self):
        super().configureView()
        self._lookup_delegate = QSqlRelationalDelegate(self._view)
        self._view.setItemDelegateForColumn(self.fieldIndex("currency_id"), self._lookup_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex("quote_source"), self._lookup_delegate)
        self._bool_delegate = BoolDelegate(self._view)
        self._view.setItemDelegateForColumn(self.fieldIndex("active"), self._bool_delegate)


# Delegate class that allows to choose data type in 'key_field' and edit data in 'value_field' (both are integer
# indices). Editors are created based on data type associated with 'key_field' via self.types dictionary
class DataDelegate(QStyledItemDelegate):    # Code doubles with pieces from delegates.py
    def __init__(self, key_field, value_field, parent=None):
        super().__init__(parent=parent)
        self._key = key_field
        self._value = value_field
        self.types = {
            AssetData.RegistrationCode: (self.tr("reg.code"), "str"),
            AssetData.ExpiryDate: (self.tr("expiry"), "date"),
            AssetData.PrincipalValue: (self.tr("principal"), "float"),
            AssetData.Tag: (self.tr("tag"), "tag")
        }

    def type(self, index):
        return self.types[index][0]

    def display_value(self, type_index, value):
        datatype = self.types[type_index][1]
        try:
            if datatype == "str":
                return value
            elif datatype == "date":
                return datetime.utcfromtimestamp(int(value)).strftime("%d/%m/%Y")
            elif datatype == "float":
                return f"{value:.2f}"
            elif datatype == "tag":
                return JalTag(int(value)).name()
            else:
                assert False, "Unknown data type of asset data"
        except ValueError:
            return ''

    def createEditor(self, aParent, option, index):
        if index.column() == self._key:
            editor = QComboBox(aParent)
            for idx in self.types:
                editor.addItem(self.types[idx][0], userData=idx)
        elif index.column() == self._value:
            type_idx = index.model().data(index.sibling(index.row(), self._key), role=Qt.EditRole)
            if self.types[type_idx][1] == "str" or self.types[type_idx][1] == "float":
                editor = QLineEdit(aParent)
            elif self.types[type_idx][1] == "date":
                editor = DateTimeEditWithReset(aParent)
                editor.setTimeSpec(Qt.UTC)
                editor.setDisplayFormat("dd/MM/yyyy")
            elif self.types[type_idx][1] == "tag":
                editor = TagSelector(aParent)
            else:
                assert False, f"Unknown data type '{self.types[type_idx][1]}' in DataDelegate.createEditor()"
        else:
            assert False, f"Delegate DataDelegate.createEditor() called for not-initialized column {index.column()}"
        return editor

    def setEditorData(self, editor, index):
        if index.column() == self._key:
            editor.setCurrentIndex(editor.findData(index.model().data(index, Qt.EditRole)))
        elif index.column() == self._value:
            type_idx = index.model().data(index.sibling(index.row(), self._key), role=Qt.EditRole)
            if self.types[type_idx][1] == "str":
                editor.setText(index.model().data(index, Qt.EditRole))
            elif self.types[type_idx][1] == "date":
                try:
                    timestamp = int(index.model().data(index, Qt.EditRole))
                    editor.setDateTime(QDateTime.fromSecsSinceEpoch(timestamp, QTimeZone(0)))
                except ValueError:
                    QStyledItemDelegate.setEditorData(self, editor, index)
            elif self.types[type_idx][1] == "float":
                try:
                    amount = Decimal(index.model().data(index, Qt.EditRole))
                except (ValueError, TypeError):
                    amount = Decimal('0')
                editor.setText(localize_decimal(amount))
            elif self.types[type_idx][1] == "tag":
                try:
                    tag_id = int(index.model().data(index, Qt.EditRole))
                except (ValueError, TypeError):
                    tag_id = 0
                editor.selected_id = tag_id
            else:
                assert False, f"Unknown data type '{self.types[type_idx][1]}' in DataDelegate.setEditorData()"
        else:
            assert False, f"Delegate DataDelegate.setEditorData() called for not-initialized column {index.column()}"

    def setModelData(self, editor, model, index):
        if index.column() == self._key:
            model.setData(index, editor.currentData())
            model.setData(index.sibling(index.row(), self._value), '')  # Reset data value on type change
        elif index.column() == self._value:
            type_idx = index.model().data(index.sibling(index.row(), self._key), role=Qt.EditRole)
            if self.types[type_idx][1] == "str":
                model.setData(index, editor.text())
            elif self.types[type_idx][1] == "date":
                timestamp = editor.dateTime().toSecsSinceEpoch()
                model.setData(index, str(timestamp))
            elif self.types[type_idx][1] == "float":
                value = QLocale().toDouble(editor.text())[0]
                model.setData(index, value)
            elif self.types[type_idx][1] == "tag":
                model.setData(index, str(editor.selected_id))
            else:
                assert False, f"Unknown data type '{self.types[type_idx][1]}' in DataDelegate.setModelData()"
        else:
            assert False, f"Delegate DataDelegate.setModelData() called for not-initialized column {index.column()}"


class ExtraDataModel(AbstractReferenceListModel):
    def __init__(self, table, parent_view):
        super().__init__(table=table, parent_view=parent_view)
        self._columns = [("id", ''),
                         ("asset_id", ''),
                         ("datatype", self.tr("Property")),
                         ("value", self.tr("Value"))]
        self._default_name = "datatype"
        self._sort_by = "datatype"
        self._hidden = ["id", "asset_id"]
        self._stretch = "value"
        self._data_delegate = None
        self._default_values = {'datatype': 1, 'value': ''}

    def configureView(self):
        super().configureView()
        self._data_delegate = DataDelegate(self.fieldIndex("datatype"), self.fieldIndex("value"), self._view)
        self._view.setItemDelegateForColumn(self.fieldIndex("datatype"), self._data_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex("value"), self._data_delegate)

    def data(self, index, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if index.column() == self.fieldIndex("datatype"):
                return self._data_delegate.type(super().data(index, role))
            elif index.column() == self.fieldIndex("value"):
                datatype = super().data(index.sibling(index.row(), self.fieldIndex("datatype")), role)
                return self._data_delegate.display_value(datatype, super().data(index, role))
        return super().data(index, role)



import logging
import re
from copy import deepcopy
from datetime import datetime
from itertools import groupby
from decimal import Decimal
from lxml import etree

from PySide6.QtWidgets import QApplication
from jal.constants import PredefinedCategory
from jal.widgets.helpers import ManipulateDate, ts2dt, ts2d
from jal.db.helpers import format_decimal
from jal.db.account import JalAccount
from jal.db.operations import Dividend
from jal.data_import.statement import FOF, Statement_ImportError, Statement_Capabilities
from jal.data_import.statement_xml import StatementXML

JAL_STATEMENT_CLASS = "StatementIBKR"
IBKR_CALCULATION_PRECISION = 10
DIVIDENDS_TABLE_ASSET_FIELD = 7

# -----------------------------------------------------------------------------------------------------------------------
class IBKRCashOp:
    Dividend = 0
    TaxWithhold = 1
    DepositWithdrawal = 2
    Fee = 3
    Interest = 4
    BondInterest = 5


# -----------------------------------------------------------------------------------------------------------------------
class IBKR_AssetType:
    NotSupported = -1
    _asset_types = {
        '': -1,
        'CASH': FOF.ASSET_MONEY,
        'STK': FOF.ASSET_STOCK,
        'ETF': FOF.ASSET_ETF,
        'ADR': FOF.ASSET_ADR,
        'BOND': FOF.ASSET_BOND,
        'OPT': FOF.ASSET_OPTION,
        'FUT': FOF.ASSET_FUTURES,
        'WAR': FOF.ASSET_WARRANT,
        'RIGHT': FOF.ASSET_RIGHTS,
        'CFD': FOF.ASSET_CFD
    }

    def __init__(self, asset_type, subtype):
        self.type = self.NotSupported
        try:
            self.type = self._asset_types[asset_type]
        except KeyError:
            raise Statement_ImportError(
                QApplication.translate("IBKR", "Asset type isn't supported: ") + f"'{asset_type}'")
        if self.type == FOF.ASSET_STOCK and subtype:  # distinguish ADR and ETF from stocks
            try:
                self.type = self._asset_types[subtype]
            except KeyError:
                pass


# -----------------------------------------------------------------------------------------------------------------------
class IBKR_CorpActionType:
    NotSupported = -1
    _corporate_action_types = {
        'BM': FOF.ACTION_BOND_MATURITY,    # No separate value as will be converted to ordinary bond sell operation
        'DW': FOF.ACTION_DELISTING,        # Delisting with loss of value
        'FS': FOF.ACTION_SPLIT,            # Forward split
        'HI': FOF.PAYMENT_STOCK_DIVIDEND,  # Choice dividend
        'IC': FOF.ACTION_SYMBOL_CHANGE,    # Issue change
        'RI': FOF.ACTION_RIGHTS_ISSUE,     # Subscribable Rights Issue
        'RS': FOF.ACTION_SPLIT,            # Reverse split
        'SO': FOF.ACTION_SPINOFF,          # Spin-off of new company
        'SD': FOF.PAYMENT_STOCK_DIVIDEND,  # Dividend paid in stocks
        'TC': FOF.ACTION_MERGER,           # Conversion of one stock into another
        'TO': FOF.ACTION_MERGER            # Voluntary conversion of one asset into another
    }

    def __init__(self, action_type):
        self.type = self.NotSupported
        try:
            self.type = self._corporate_action_types[action_type]
        except KeyError:
            raise Statement_ImportError(
                QApplication.translate("IBKR", "Corporate action isn't supported: ") + f"{action_type}")


# -----------------------------------------------------------------------------------------------------------------------
class IBKR_Currency:
    pass

# -----------------------------------------------------------------------------------------------------------------------
class IBKR_Asset:
    BondPrincipal = 1000


# -----------------------------------------------------------------------------------------------------------------------
class IBKR_Account:
    def __init__(self, accounts_list, number, currency_ids):
        self.id = None
        account_ids = []
        for currency in currency_ids:
            match = [x for x in accounts_list if x['number'] == number and x['currency'] == currency]
            if match:
                if len(match) == 1:
                    account_ids.append(match[0]["id"])
                else:
                    logging.error(QApplication.translate("IBKR", "Multiple account match for ") + f"{number}")
            else:
                new_id = max([0] + [x['id'] for x in accounts_list]) + 1
                account_ids.append(new_id)
                account = {"id": new_id, "number": number,
                           "currency": currency, "precision": IBKR_CALCULATION_PRECISION}
                accounts_list.append(account)
        if account_ids:
            if len(account_ids) == 1:
                self.id = account_ids[0]
            else:
                self.id = account_ids


# -----------------------------------------------------------------------------------------------------------------------
# Class for Loading Interactive Brokers XML Flex report
class StatementIBKR(StatementXML):
    statements_path = '/*/FlexStatement'
    statement_tag = 'FlexStatement'
    level_tag = 'levelOfDetail'
    CancelledFlag = 'Ca'
    ReversalCode = 'Re'
    ReversalSuffix = " - REVERSAL"
    CancelPrefix = "CANCEL "
    NoAsset = -1

    def __init__(self):
        super().__init__()
        self.name = self.tr("Interactive Brokers")
        self.icon_name = "ibkr.png"
        self.filename_filter = self.tr("IBKR flex-query (*.xml)")

        ibkr_loaders = {
            IBKR_Currency: self.attr_currency,
            IBKR_AssetType: self.attr_asset_type,
            IBKR_Asset: self.attr_asset,
            IBKR_Account: self.attr_account,
            IBKR_CorpActionType: self.attr_corp_action_type
        }
        self.attr_loader.update(ibkr_loaders)
        self._sections = {   # Order of load is important - accounts and assets should be loaded first
            StatementXML.STATEMENT_ROOT: {'tag': self.statement_tag,
                                          'level': '',
                                          'values': [('accountId', 'account', str, None),
                                                     ('fromDate', 'period_start', datetime, None),
                                                     ('toDate', 'period_end', datetime, None)],
                                          'loader': self.load_header
                                          },
            'CashReport': {'tag': 'CashReportCurrency',
                           'level': 'Currency',
                           'values': [('accountId', 'number', str, None),
                                      ('currency', 'currency', IBKR_Currency, None),
                                      ('startingCash', 'cash_begin', float, None),
                                      ('endingCash', 'cash_end', float, None),                   # -- this is planned
                                      ('endingSettledCash', 'cash_end_settled', float, None)],   # -- this is now
                           'loader': self.load_accounts},
            'SecuritiesInfo': {'tag': 'SecurityInfo',
                               'level': '',
                               'values': [('symbol', 'symbol', str, None),
                                          ('currency', 'currency', IBKR_Currency, None),
                                          ('assetCategory', 'type', IBKR_AssetType, IBKR_AssetType.NotSupported),
                                          ('description', 'name', str, None),
                                          ('isin', 'isin', str, ''),
                                          ('cusip', 'reg_number', str, ''),
                                          ('expiry', 'expiry', datetime, 0),
                                          ('maturity', 'maturity', datetime, 0),
                                          ('principalAdjustFactor', 'principal', str, ''),
                                          ('listingExchange', 'exchange', str, '')],
                               'loader': self.load_assets},
            'Trades': {'tag': 'Trade',
                       'level': 'EXECUTION',
                       'values': [('assetCategory', 'type', IBKR_AssetType, IBKR_AssetType.NotSupported),
                                  ('symbol', 'asset', IBKR_Asset, None),
                                  ('accountId', 'account', IBKR_Account, None),
                                  ('dateTime', 'timestamp', datetime, None),
                                  ('settleDateTarget', 'settlement', datetime, 0),
                                  ('tradePrice', 'price', float, None),
                                  ('quantity', 'quantity', float, None),
                                  ('proceeds', 'proceeds', float, None),
                                  ('multiplier', 'multiplier', float, None),
                                  ('ibCommission', 'fee', float, None),
                                  ('tradeID', 'number', str, ''),
                                  ('exchange', 'exchange', str, ''),
                                  ('notes', 'notes', str, '')],
                       'loader': self.load_ib_trades},
            'OptionEAE': {'tag': 'OptionEAE',
                          'level': '',
                          'values': [('transactionType', 'operation', str, None),
                                     ('symbol', 'asset', IBKR_Asset, None),
                                     ('accountId', 'account', IBKR_Account, None),
                                     ('date', 'timestamp', datetime, None),
                                     ('tradePrice', 'price', float, None),
                                     ('quantity', 'quantity', float, None),
                                     ('multiplier', 'multiplier', float, None),
                                     ('commisionsAndTax', 'fee', float, None),
                                     ('tradeID', 'number', str, ''),
                                     ('notes', 'notes', str, '')],
                          'loader': self.load_options},
            'CorporateActions': {'tag': 'CorporateAction',
                                 'level': 'DETAIL',
                                 'values': [('type', 'type', IBKR_CorpActionType, IBKR_CorpActionType.NotSupported),
                                            ('accountId', 'account', IBKR_Account, None),
                                            ('symbol', 'asset', IBKR_Asset, None),
                                            ('assetCategory', 'asset_type', IBKR_AssetType, IBKR_AssetType.NotSupported),
                                            ('dateTime', 'timestamp', datetime, None),
                                            ('transactionID', 'number', str, ''),
                                            ('description', 'description', str, None),
                                            ('quantity', 'quantity', float, None),
                                            ('value', 'value', float, None),
                                            ('proceeds', 'proceeds', float, None),
                                            ('code', 'code', str, '')],
                                 'loader': self.load_corporate_actions},
            'CashTransactions': {'tag': 'CashTransaction',
                                 'level': 'DETAIL',
                                 'values': [('type', 'type', str, None),
                                            ('accountId', 'account', IBKR_Account, None),
                                            ('symbol', 'asset', IBKR_Asset, 0),
                                            ('currency', 'currency', IBKR_Currency, None),
                                            ('dateTime', 'timestamp', datetime, None),
                                            ('reportDate', 'reported', datetime, None),
                                            ('amount', 'amount', float, None),
                                            ('tradeID', 'number', str, ''),
                                            ('transactionID', 'tid', str, ''),
                                            ('description', 'description', str, None)],
                                 'loader': self.load_cash_transactions},
            'ChangeInDividendAccruals': {'tag': 'ChangeInDividendAccrual',
                                         'level': 'DETAIL',
                                         'values': [('code', 'code', str, None),
                                                    ('accountId', 'account', IBKR_Account, None),
                                                    ('symbol', 'asset', IBKR_Asset, 0),
                                                    ('currency', 'currency', IBKR_Currency, None),
                                                    ('date', 'timestamp', datetime, None),
                                                    ('exDate', 'ex_date', datetime, None),
                                                    ('grossAmount', 'amount', float, None),
                                                    ('grossRate', 'rate', float, 0)],
                                         'loader': self.load_dividend_accruals},
            'StockGrantActivities': {'tag': 'StockGrantActivity',
                                     'level': '',
                                     'values': [('accountId', 'account', IBKR_Account, None),
                                                ('symbol', 'asset', IBKR_Asset, None),
                                                ('awardDate', 'timestamp', datetime, None),
                                                ('activityDescription', 'description', str, None),
                                                ('quantity', 'amount', float, None),
                                                ('price', 'price', str, None)],
                                     'loader': self.load_vestings},
            'TransactionTaxes': {'tag': 'TransactionTax',
                                 'level': 'SUMMARY',
                                 'values': [('accountId', 'account', IBKR_Account, None),
                                            ('symbol', 'symbol', str, None),
                                            ('date', 'timestamp', datetime, None),
                                            ('taxAmount', 'amount', float, None),
                                            ('description', 'description', str, None),
                                            ('taxDescription', 'tax_description', str, None)],
                                 'loader': self.load_taxes},
            'CFDCharges': {'tag': 'CFDCharge',
                           'level': '',
                           'values': [('accountId', 'account', IBKR_Account, None),
                                      ('symbol', 'asset', IBKR_Asset, self.NoAsset),
                                      ('date', 'timestamp', datetime, None),
                                      ('total', 'amount', float, None),
                                      ('transactionID', 'number', str, ''),
                                      ('activityDescription', 'description', str, None)],
                           'loader': self.load_cfd_charges},
        }

    @staticmethod
    def tr(text):
        return QApplication.translate("StatementIBKR", text)

    @staticmethod
    def capabilities() -> set:
        return {Statement_Capabilities.MULTIPLE_LOAD}

    # Saves information from XML that is related with a given asset
    def save_debug_info(self, account, asset):
        # Dump statement info relevant to given asset
        debug_info = 'Statement data:\n----------------------------------------------------------------\n'
        assert self._statement is not None
        symbols = [x['symbol'] for x in self._data[FOF.SYMBOLS] if x["asset"] == asset]
        for symbol in symbols:
            elements = self._statement.findall(f".//*[@symbol='{symbol}']")
            for element in elements:
                if 'accountId' in element.attrib:
                    element.attrib['accountId'] = 'U7654321'  # Hide real account number
                debug_info += etree.tostring(element).decode("utf-8")
        debug_info += "----------------------------------------------------------------\n"
        # Dump dividends info from database for the given asset from
        db_account = self._map_db_account(account)
        db_asset = self._map_db_asset(asset)
        dividends = JalAccount(db_account).dump_dividends()
        dividends = [x for x in dividends if x[DIVIDENDS_TABLE_ASSET_FIELD] == db_asset]
        debug_info += "Database data:\n----------------------------------------------------------------\n"
        debug_info += str(dividends)
        debug_info += "\n----------------------------------------------------------------\n"
        super().save_debug_info(debug_info=debug_info)

    def validate_file_header_attributes(self, attributes):
        if 'type' not in attributes:
            raise Statement_ImportError(self.tr("Interactive Brokers report type not found"))
        if attributes['type'] == "TCF":
            raise Statement_ImportError(self.tr("You try to import Trade confimation report, not Activity report"))
        if attributes['type'] != 'AF':
            raise Statement_ImportError(self.tr("Unknown Interactive Brokers report type: ") + f"{attributes['type']}")

    # Convert attribute 'attr_name' value into json open-format asset type
    @staticmethod
    def attr_asset_type(xml_element, attr_name, default_value):
        if attr_name not in xml_element.attrib:
            return default_value
        sub_type = xml_element.attrib['subCategory'] if 'subCategory' in xml_element.attrib else ''
        return IBKR_AssetType(xml_element.attrib[attr_name], sub_type).type

    # Convert attribute 'attr_name' value into JAL corporate action
    def attr_corp_action_type(self, xml_element, attr_name, default_value):
        if attr_name not in xml_element.attrib:
            return default_value
        asset_category = self.attr_asset_type(xml_element, 'assetCategory', None)
        return IBKR_CorpActionType(xml_element.attrib[attr_name]).type

    def attr_currency(self, xml_element, attr_name, default_value):
        if attr_name not in xml_element.attrib:
            return default_value
        currency_id = self.currency_id(xml_element.attrib[attr_name])
        if currency_id is None:
            return default_value
        else:
            return currency_id

    def attr_asset(self, xml_element, attr_name, default_value):
        if attr_name not in xml_element.attrib:
            return default_value
        if xml_element.attrib[attr_name] == '':
            return default_value
        if xml_element.tag == 'CashTransaction' and attr_name == 'currency':
            asset_category = FOF.ASSET_MONEY
        else:
            asset_category = self.attr_asset_type(xml_element, 'assetCategory', None)
        if xml_element.tag == 'Trade' and asset_category == FOF.ASSET_MONEY:
            currency = xml_element.attrib[attr_name].split('.')
            asset_id = [self.currency_id(code) for code in currency]
            if not asset_id:
                return default_value
        else:
            symbol = xml_element.attrib[attr_name]
            if symbol.endswith('.OLD'):
                symbol = symbol[:-len('.OLD')]
            if asset_category == IBKR_AssetType.NotSupported:
                raise Statement_ImportError(self.tr("Asset type isn't supported: ") + f"'{asset_category}' ({symbol})")
            asset_data = {'symbol': symbol, 'type': asset_category}
            if xml_element.tag not in ['CorporateAction', 'CashTransaction'] and 'description' in xml_element.attrib:
                asset_data['name'] = xml_element.attrib['description']
            if asset_category != FOF.ASSET_MONEY:
                asset_data['currency'] = self.currency_id(xml_element.attrib['currency'])
            if 'isin' in xml_element.attrib and xml_element.attrib['isin']:
                asset_data['isin'] = xml_element.attrib['isin']
            if 'cusip' in xml_element.attrib and xml_element.attrib['cusip']:
                asset_data['reg_number'] = xml_element.attrib['cusip']
            if 'listingExchange' in xml_element.attrib and xml_element.attrib['listingExchange'] \
                    and xml_element.attrib['listingExchange'] != 'VALUE':  # don't store 'VALUE' or empty exchange
                asset_data['note'] = xml_element.attrib['listingExchange']
            asset_id = self.asset_id(asset_data)
        return asset_id

    def attr_account(self, xml_element, attr_name, default_value):
        if attr_name not in xml_element.attrib:
            return default_value
        if xml_element.tag == 'Trade' and self.attr_asset_type(xml_element, 'assetCategory', None) == FOF.ASSET_MONEY:
            if 'symbol' not in xml_element.attrib or 'ibCommissionCurrency' not in xml_element.attrib:
                if default_value is None:
                    logging.error(self.tr("Can't get currencies for currency exchange: ") + f"{xml_element}")
                return default_value
            currency = xml_element.attrib['symbol'].split('.')
            currency.append(xml_element.attrib['ibCommissionCurrency'])
        else:
            if 'currency' not in xml_element.attrib:
                if default_value is None:
                    logging.error(self.tr("Can't get account currency for account: ") + f"{xml_element}")
                return default_value
            currency = [xml_element.attrib['currency']]
        currency_ids = [self.currency_id(code) for code in currency]
        account_id = IBKR_Account(self._data[FOF.ACCOUNTS], xml_element.attrib[attr_name], currency_ids).id
        if account_id is None:
            return default_value
        else:
            return account_id

    def locate_asset(self, symbol, isin) -> int:
        candidates = [x for x in self._data[FOF.ASSETS] if 'isin' in x and x['isin'] == isin]
        if len(candidates) == 1:
            return candidates[0]["id"]
        candidates = [x for x in self._data[FOF.SYMBOLS] if 'symbol' in x and x['symbol'] == symbol]
        if len(candidates) == 1:
            return candidates[0]["asset"]
        return 0

    def set_asset_counry(self, asset_id, country):
        assets = [x for x in self._data[FOF.ASSETS] if 'id' in x and x['id'] == asset_id]
        if len(assets) != 1:
            return
        assets[0]["country"] = country

    def load_header(self, header):
        self._data[FOF.PERIOD][0] = header['period_start']
        self._data[FOF.PERIOD][1] = self._end_of_date(header['period_end'])
        logging.info(self.tr("Load IB Flex-statement for account ") +
                     f"{header['account']}: {datetime.utcfromtimestamp(header['period_start']).strftime('%Y-%m-%d')}" +
                     f" - {datetime.utcfromtimestamp(header['period_end']).strftime('%Y-%m-%d')}")

    def load_accounts(self, balances):
        for i, balance in enumerate(sorted(balances, key=lambda x: x['currency'])):
            balance['id'] = i + 1
            balance['precision'] = IBKR_CALCULATION_PRECISION
            self._data[FOF.ACCOUNTS].append(balance)

    def load_assets(self, assets):
        asset_count = 0
        for asset in assets:
            if asset['type'] == IBKR_AssetType.NotSupported:   # Skip not supported type of asset
                continue
            # IB may use '.OLD' suffix if asset is being replaced
            asset['symbol'] = asset['symbol'][:-len('.OLD')] if asset['symbol'].endswith('.OLD') else asset['symbol']
            if asset['exchange'] and asset['exchange'] != 'VALUE':  # don't store 'VALUE' or empty exchange
                asset['note'] = asset['exchange']
            if asset['maturity']:
                asset['expiry'] = asset['maturity']
            if asset['expiry'] == 0:
                asset.pop('expiry')
            if asset['principal']:
                asset['principal'] = int(asset['principal']) * IBKR_Asset.BondPrincipal
            asset.pop('maturity')
            asset.pop('exchange')
            self.asset_id(asset)
            asset_count += 1
        logging.info(self.tr("Securities loaded: ") + f"{asset_count} ({len(assets)})")

    def load_ib_trades(self, ib_trades):
        trades = [trade for trade in ib_trades if type(trade['asset']) == int]
        trades_loaded = self.load_trades(trades)

        transfers = [transfer for transfer in ib_trades if type(transfer['asset']) == list]
        transfers_loaded = self.load_transfers(transfers)

        logging.info(self.tr("Trades loaded: ") + f"{trades_loaded + transfers_loaded} ({len(ib_trades)})")

    def load_trades(self, trades):
        trade_base = max([0] + [x['id'] for x in self._data[FOF.TRADES]]) + 1
        cnt = 0
        for i, trade in enumerate(sorted(trades, key=lambda x: x['timestamp'])):
            trade['id'] = trade_base + i
            trade['quantity'] = trade['quantity'] * trade['multiplier']
            if trade['settlement'] == 0:
                trade['settlement'] = trade['timestamp']
            asset = [x for x in self._data[FOF.ASSETS] if x['id'] == trade['asset']][0]
            if asset['type'] == FOF.ASSET_BOND:
                trade['quantity'] = trade['quantity'] / IBKR_Asset.BondPrincipal
                trade['price'] = trade['price'] * IBKR_Asset.BondPrincipal / 100.0  # Bonds are priced in percents of principal
            trade['fee'] = -trade['fee'] if trade['fee'] != 0 else 0.0  # otherwise we may have negative 0.0
            if trade['notes'] == StatementIBKR.CancelledFlag:
                trade['cancelled'] = True
            self.drop_extra_fields(trade, ["type", "proceeds", "multiplier", "exchange", "notes"])
            self._data[FOF.TRADES].append(trade)
            cnt += 1
        return cnt

    def load_transfers(self, transfers):
        transfer_base = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        cnt = 0
        for i, transfer in enumerate(sorted(transfers, key=lambda x: x['timestamp'])):
            transfer['id'] = transfer_base + i
            if transfer['quantity'] > 0:
                transfer['account'][0], transfer['account'][1] = transfer['account'][1], transfer['account'][0]
                transfer['asset'][0], transfer['asset'][1] = transfer['asset'][1], transfer['asset'][0]
                transfer['quantity'], transfer['proceeds'] = transfer['proceeds'], transfer['quantity']
            transfer['withdrawal'] = abs(transfer.pop('quantity'))
            transfer['deposit'] = abs(transfer.pop('proceeds'))
            transfer['fee'] = -transfer['fee'] if transfer['fee'] != 0 else 0.0  # otherwise we may have negative 0.0
            transfer['description'] = transfer['exchange']
            self.drop_extra_fields(transfer, ["type", "settlement", "price", "multiplier", "exchange", "notes"])
            self._data[FOF.TRANSFERS].append(transfer)
            cnt += 1
        return cnt

    def load_options(self, options):
        transaction_desctiption = {
            "Assignment": self.tr("Option assignment"),
            "Exercise": self.tr("Option exercise"),
            "Expiration": self.tr("Option expiration"),
            "Buy": self.tr("Option assignment/exercise"),
            "Sell": self.tr("Option assignment/exercise"),
        }
        cnt = 0
        for option in options:
            description = ''
            try:
                description = transaction_desctiption[option['operation']]
            except KeyError:
                logging.error(
                    self.tr("Option E&A&E action isn't implemented: ") + f"{option['transactionType']}")
            if description:
                trade = [x for x in self._data[FOF.TRADES] if x['account'] == option['account']
                         and x['asset'] == option['asset'] and x['number'] == option['number']]
                if len(trade) == 1:
                    trade[0]['note'] = description
                else:
                    raise Statement_ImportError(
                        self.tr("Original trade not found for Option E&A&E operation: ") + f"{option}")
                cnt += 1
        logging.info(self.tr("Options E&A&E loaded: ") + f"{cnt} ({len(options)})")

    def load_corporate_actions(self, actions):
        action_loaders = {
            FOF.ACTION_MERGER: self.load_merger,
            FOF.ACTION_SPINOFF: self.load_spinoff,
            FOF.ACTION_SYMBOL_CHANGE: self.load_symbol_change,
            FOF.PAYMENT_STOCK_DIVIDEND: self.load_stock_dividend,
            FOF.ACTION_SPLIT: self.load_split,
            FOF.ACTION_BOND_MATURITY: self.load_bond_maturity,
            FOF.ACTION_DELISTING: self.load_delisting,
            FOF.ACTION_RIGHTS_ISSUE: self.load_none
        }

        cnt = 0
        self.remove_cancelled_corporate_actions(actions)

        # If stocks were bought/sold on a corporate action day IBKR may put several records for one corporate
        # action. So first step is to aggregate quantity.
        key_func = lambda x: (x['account'], x['asset'], x['type'], x['description'], x['timestamp'])
        actions_sorted = sorted(actions, key=key_func)
        actions_aggregated = []
        for k, group in groupby(actions_sorted, key=key_func):
            group_list = list(group)
            part = group_list[0]  # Take fist of several actions as a basis
            part['quantity'] = sum(action['quantity'] for action in group_list)  # and update quantity in it
            part['jal_processed'] = False  # This flag will be used to mark already processed records
            actions_aggregated.append(part)
            cnt += len(group_list) - 1
        # Now split in 2 parts: A for new stocks deposit, B for old stocks withdrawal
        # There might be 0 quantity value - it should be ignored
        parts_a = [action for action in actions_aggregated if action['quantity'] > 0]
        parts_b = [action for action in actions_aggregated if action['quantity'] < 0]
        # Process sequentially '+' and '-', 'jal_processed' will set True when '+' has pair record in '-'
        for action in parts_a + parts_b:
            if action['jal_processed']:
                continue
            if action['type'] in action_loaders:
                cnt += action_loaders[action['type']](action, parts_b)
            else:
                raise Statement_ImportError(
                    self.tr("Corporate action type is not supported: ") + f"{action}")
        logging.info(self.tr("Corporate actions loaded: ") + f"{cnt} ({len(actions)})")

    # Find record in list 'parts_b' (second parts of corporate actions) which matches
    # given asset and description with more details from corp_action itself
    def find_corp_action_pair(self, asset, description, action, parts_b):
        paired_record = list(filter(
            lambda pair: pair['asset'] == asset
                         and (pair['description'].startswith(description + ", ")
                              or pair['description'].startswith(description + ".OLD, "))
                         and pair['type'] == action['type']
                         and pair['timestamp'] == action['timestamp'], parts_b))
        if len(paired_record) != 1:
            raise Statement_ImportError(self.tr("Can't find paired record for ") + f"{action}")
        return paired_record

    # Takes cancelled corporate actions and tries to find and remove original one form actions list
    def remove_cancelled_corporate_actions(self, actions):
        delete_elements = []
        cancelled_actions = [(idx, action) for idx, action in enumerate(actions) if action['code'] == StatementIBKR.CancelledFlag]
        for c_id, c_action in cancelled_actions:
            matched = [idx for (idx, action) in enumerate(actions) if
                       action["type"] == c_action["type"] and action["account"] == c_action["account"] and
                       action["asset"] == c_action["asset"] and action["timestamp"] == c_action["timestamp"] and
                       action["description"] == c_action["description"] and action["quantity"] == -c_action["quantity"]]
            if len(matched) == 1:
                delete_elements += [c_id, matched[0]]
        for idx in sorted(delete_elements, reverse=True):
            del actions[idx]
        for action in actions:
            if action['code'] == StatementIBKR.CancelledFlag:
                raise Statement_ImportError(self.tr("Can't process cancelled corporate action") + f" '{action}'")

    # Dummy loader to skip some corporate actions
    def load_none(self, action, parts_b) -> int:
        return 0

    def load_merger(self, action, parts_b) -> int:
        MergerPatterns = [
            r"^(?P<symbol_old>.*)(.OLD)?\((?P<isin_old>\w+)\) +MERGED\([\w ]+\) +WITH +(?P<isin_new>\w+) +(?P<X>\d+) +FOR +(?P<Y>\d+) +\((?P<symbol>.*)(.OLD)?, (?P<name>.*), (?P<id>\w+)\)$",
            r"^(?P<symbol_old>.*)(.OLD)?\((?P<isin_old>\w+)\) +CASH and STOCK MERGER +\([\w ]+\) +(?P<isin_new>[\w ]+) +(?P<X>\d+) +FOR +(?P<Y>\d+) +AND +(?P<currency>\w+) +(\d+(\.\d+)?) +\((?P<symbol>[\w ]+)(.OLD)?, (?P<name>.*), (?P<id>\w+)\)$",
            r"^(?P<symbol_old>.*)(.OLD)?\((?P<isin_old>\w+)\) +CASH and STOCK MERGER +\([\w ]+\) +(?P<isin_new>[\w ]+) +(?P<X>\d+) +FOR +(?P<Y>\d+), +(?P<isin_new2>[\w ]+) +(?P<X2>\d+) +FOR +(?P<Y2>\d+) +AND +(?P<currency>\w+) +(\d+(\.\d+)?) +\((?P<symbol>[\w ]+)(.OLD)?, (?P<name>.*), (?P<id>\w+)\)$",
            r"^(?P<symbol_old>.*)\((?P<isin_old>\w+)\) +TENDERED TO +(?P<isin_new>\w+) +(?P<X>\d+) +FOR +(?P<Y>\d+) +\((?P<symbol>.*), +(?P<name>.*), +(?P<id>.*)\)$",
            r"^(?P<symbol_old>.*)\((?P<isin_old>\w+)\) +MERGED\([\w ]+\) +FOR (?P<currency>\w+) (?P<price>\d+\.\d+) PER SHARE +\((?P<symbol>.*), (?P<name>.*)( - TENDER ODD LOT)?, (?P<id>\w+)\)$",  # "TENDER ODD LOT" part is optional
            r"^(?P<symbol_old>.*)(.OLD)?\((?P<isin_old>\w+)\) +MERGED\([\w ]+\) +WITH +(?P<isin_new>.*) +(?P<X>\d+) +FOR +(?P<Y>\d+), +(?P<isin_new2>.*) +(?P<X2>\d+) +FOR +(?P<Y2>\d+) +\((?P<symbol>.*)(.OLD)?, (?P<name>.*), (?P<id>\w+)\)$"
            ]

        parts = None
        pattern_id = -1
        for pattern_id, pattern in enumerate(MergerPatterns):
            parts = re.match(pattern, action['description'], re.IGNORECASE)
            if parts:
                break

        if parts is None:
            raise Statement_ImportError(self.tr("Can't parse Merger description ") + f"'{action}'")
        merger_a = parts.groupdict()

        if len(merger_a) != MergerPatterns[pattern_id].count("(?P<"):  # check expected number of matches
            raise Statement_ImportError(self.tr("Merger description miss some data ") + f"'{action}'")

        description_b = action['description'][:parts.span('symbol')[0]] + merger_a['symbol_old']
        asset_b = self.locate_asset(merger_a['symbol_old'], merger_a['isin_old'])

        if pattern_id == 4:  # Asset converted to money -> store it as a sell trade
            action['id'] = max([0] + [x['id'] for x in self._data[FOF.TRADES]]) + 1
            action['settlement'] = action['timestamp']
            action['price'] = action['proceeds'] / (-action['quantity'])
            action['note'] = action.pop('description')
            action['fee'] = 0.0
            self.drop_extra_fields(action, ["type", "value", "proceeds", "code", "asset_type", "jal_processed"])
            self._data[FOF.TRADES].append(action)
            return 1

        paired_record = self.find_corp_action_pair(asset_b, description_b, action, parts_b)
        # Adjust quantity for bonds
        adj_factor = IBKR_Asset.BondPrincipal if action['asset_type'] == FOF.ASSET_BOND else 1.0
        existing_action = None
        # Special processing if 1 asset is converted into two other assets
        if pattern_id == 2 or pattern_id == 5:
            existing_action = self.locate_existing_merger(action['timestamp'],
                                                          action['account'], paired_record[0]['asset'])
        if existing_action is None:
            action['id'] = max([0] + [x['id'] for x in self._data[FOF.CORP_ACTIONS]]) + 1
            action['outcome'] = [{'asset': action['asset'], 'quantity': action['quantity']/adj_factor, 'share': 0.0}]
            action['asset'] = paired_record[0]['asset']
            action['quantity'] = -paired_record[0]['quantity']/adj_factor
            # Process cash payment if it is present as part of corporate action
            if pattern_id == 1 or pattern_id == 2:
                payment = {'asset': self.currency_id(parts['currency']),
                           'quantity': paired_record[0]['proceeds'], 'share': 0.0}
                action['outcome'].insert(0, payment)
            self.drop_extra_fields(action, ["value", "proceeds", "code", "asset_type", "jal_processed"])
            self._data[FOF.CORP_ACTIONS].append(action)
        else:
            next_outcome = {'asset': action['asset'], 'quantity': action['quantity']/adj_factor, 'share': 0.0}
            existing_action['outcome'].append(next_outcome)
        paired_record[0]['jal_processed'] = True
        return 2

    def locate_existing_merger(self, timestamp, account, asset):
        existing_merger = list(filter(
            lambda merger: merger['type'] == FOF.ACTION_MERGER and merger['timestamp'] == timestamp
                           and merger['account'] == account and merger['asset'] == asset, self._data[FOF.CORP_ACTIONS]))
        if len(existing_merger) == 0:
            return None
        if len(existing_merger) != 1:
            raise Statement_ImportError(self.tr("Multiple merger records already exist at ") + f"{timestamp}")
        return existing_merger[0]

    def load_spinoff(self, action, _parts_b) -> int:
        SpinOffPattern = r"^(?P<symbol_old>.*)\((?P<isin_old>\w+)\) +SPINOFF +(?P<X>\d+) +FOR +(?P<Y>\d+) +\((?P<symbol>.*), (?P<name>.*), (?P<id>\w+)\)$"

        parts = re.match(SpinOffPattern, action['description'], re.IGNORECASE)
        if parts is None:
            raise Statement_ImportError(self.tr("Can't parse Spin-off description ") + f"'{action}'")
        spinoff = parts.groupdict()
        if len(spinoff) != SpinOffPattern.count("(?P<"):  # check that expected number of groups was matched
            raise Statement_ImportError(self.tr("Spin-off description miss some data ") + f"'{action}'")
        asset_old = self.locate_asset(spinoff['symbol_old'], spinoff['isin_old'])
        if not asset_old:
            raise Statement_ImportError(self.tr("Spin-off initial asset not found ") + f"'{action}'")
        qty_old = int(spinoff['Y']) * action['quantity'] / int(spinoff['X'])
        if abs(round(qty_old) - qty_old) > 0.01:
            raise Statement_ImportError(self.tr("Spin-off rounding error is too big ") + f"'{action}'")
        qty_old = round(qty_old)
        action['id'] = max([0] + [x['id'] for x in self._data[FOF.CORP_ACTIONS]]) + 1
        action['outcome'] = [{'asset': asset_old, 'quantity': qty_old, 'share': 0.0},
                             {'asset': action['asset'], 'quantity': action['quantity'], 'share': 0.0}]
        action['asset'] = asset_old
        action['quantity'] = qty_old
        self.drop_extra_fields(action, ["value", "proceeds", "code", "asset_type", "jal_processed"])
        self._data[FOF.CORP_ACTIONS].append(action)
        return 1

    def load_symbol_change(self, action, parts_b) -> int:
        SymbolChangePattern = r"^(?P<symbol_old>.*)\((?P<isin_old>\w+)\) +CUSIP\/ISIN CHANGE TO +\((?P<isin_new>\w+)\) +\((?P<symbol>.*), (?P<name>.*), (?P<id>\w+)\)$"

        parts = re.match(SymbolChangePattern, action['description'], re.IGNORECASE)
        if parts is None:
            raise Statement_ImportError(self.tr("Can't parse Symbol Change description ") + f"'{action}'")
        isin_change = parts.groupdict()
        if len(isin_change) != SymbolChangePattern.count("(?P<"):  # check that expected number of groups was matched
            raise Statement_ImportError(self.tr("Spin-off description miss some data ") + f"'{action}'")
        description_b = action['description'][:parts.span('symbol')[0]] + isin_change['symbol_old']
        asset_b = self.locate_asset(isin_change['symbol_old'], isin_change['isin_old'])
        paired_record = self.find_corp_action_pair(asset_b, description_b, action, parts_b)
        action['id'] = max([0] + [x['id'] for x in self._data[FOF.CORP_ACTIONS]]) + 1
        action['outcome'] = [{'asset': action['asset'], 'quantity': action['quantity'], 'share': 1.0}]
        action['asset'] = paired_record[0]['asset']
        action['quantity'] = -paired_record[0]['quantity']
        self.drop_extra_fields(action, ["value", "proceeds", "code", "asset_type", "jal_processed"])
        self._data[FOF.CORP_ACTIONS].append(action)
        paired_record[0]['jal_processed'] = True
        return 2

    def load_stock_dividend(self, action, parts_b) -> int:
        StockDividendPattern = r"^(?P<description>.*) +(?P<tail>\(.*\))$"

        parts = re.match(StockDividendPattern, action['description'], re.IGNORECASE)
        if parts is None:
            raise Statement_ImportError(self.tr("Can't parse Stock Dividend description ") + f"'{action}'")
        action['description'] = parts.groupdict()['description']

        action['id'] = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
        action['amount'] = action['quantity']
        action['price'] = format_decimal(Decimal(str(action['value'])) / Decimal(str(action['quantity'])))
        action['tax'] = 0
        self.drop_extra_fields(action, ["quantity", "value", "proceeds", "code", "asset_type", "jal_processed"])
        self._data[FOF.ASSET_PAYMENTS].append(action)
        return 1

    def load_split(self, action, parts_b) -> int:
        SplitPattern = r"^(?P<symbol_old>.*)\((?P<isin_old>\w+)\) +SPLIT +(?P<X>\d+) +FOR +(?P<Y>\d+) +\((?P<symbol>.*), (?P<name>.*), (?P<id>\w+)*\)$"

        parts = re.match(SplitPattern, action['description'], re.IGNORECASE)
        if parts is None:
            raise Statement_ImportError(self.tr("Can't parse Split description ") + f"'{action}'")
        split = parts.groupdict()
        if len(split) != SplitPattern.count("(?P<"):  # check that expected number of groups was matched
            raise Statement_ImportError(self.tr("Split description miss some data ") + f"'{action}'")
        if parts['id'] is None or parts['isin_old'] == parts['id']:  # Simple split without ISIN change
            qty_delta = action['quantity']
            qty_old = qty_delta / (int(split['X']) / int(split['Y']) - 1)
            qty_new = qty_old + qty_delta
            action['id'] = max([0] + [x['id'] for x in self._data[FOF.CORP_ACTIONS]]) + 1
            action['outcome'] = [{'asset': action['asset'], 'quantity': qty_new, 'share': 1.0}]
            action['quantity'] = qty_old
            self.drop_extra_fields(action, ["value", "proceeds", "code", "asset_type", "jal_processed"])
            self._data[FOF.CORP_ACTIONS].append(action)
            return 1
        else:  # Split together with ISIN change and there should be 2nd record available
            description_b = action['description'][:parts.span('symbol')[0]] + split['symbol_old']
            asset_b = self.locate_asset(split['symbol_old'], split['isin_old'])
            paired_record = self.find_corp_action_pair(asset_b, description_b, action, parts_b)
            action['id'] = max([0] + [x['id'] for x in self._data[FOF.CORP_ACTIONS]]) + 1
            action['outcome'] = [{'asset': action['asset'], 'quantity': action['quantity'], 'share': 1.0}]
            action['asset'] = paired_record[0]['asset']
            action['quantity'] = -paired_record[0]['quantity']
            self.drop_extra_fields(action, ["value", "proceeds", "code", "asset_type", "jal_processed"])
            self._data[FOF.CORP_ACTIONS].append(action)
            paired_record[0]['jal_processed'] = True
            return 2

    # Bond maturity is processed as ordinary bond
    def load_bond_maturity(self, action, parts_b) -> int:
        action['id'] = max([0] + [x['id'] for x in self._data[FOF.TRADES]]) + 1
        action['quantity'] = action['quantity'] / IBKR_Asset.BondPrincipal
        action['price'] = action['proceeds'] / (-action['quantity'])  # Quantity is negative, bonds are withdrawn
        action['settlement'] = action['timestamp']                    # Settled by the same date
        action['note'] = action['description']
        action['fee'] = 0.0
        self.drop_extra_fields(action, ["description", "value", "proceeds", "type", "code", "asset_type",
                                        "jal_processed"])
        self._data[FOF.TRADES].append(action)
        return 1

    def load_delisting(self, action, parts_b) -> int:
        # There might be delisting for issued rights - we don't need to store it as it isn't a real asset
        asset = [x for x in self._data[FOF.ASSETS] if x['id'] == action['asset']][0]
        if asset['type'] == FOF.ASSET_RIGHTS:
            return 0
        action['id'] = max([0] + [x['id'] for x in self._data[FOF.CORP_ACTIONS]]) + 1
        action['asset'] = action['asset']
        action['quantity'] = -action['quantity']
        action['outcome'] = []
        self.drop_extra_fields(action, ["value", "proceeds", "code", "asset_type", "jal_processed"])
        self._data[FOF.CORP_ACTIONS].append(action)
        return 1

    def load_vestings(self, vestings):
        cnt = 0
        asset_payments_base = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
        for i, vesting in enumerate(vestings):
            vesting['id'] = asset_payments_base + i
            vesting['type'] = FOF.PAYMENT_STOCK_VESTING
            self._data[FOF.ASSET_PAYMENTS].append(vesting)
            cnt += 1
        logging.info(self.tr("Stock vestings loaded: ") + f"{cnt} ({len(vestings)})")

    def load_cash_transactions(self, cash):
        drop_fields = lambda x, y: {i: x[i] for i in x if i not in y}  # removes from dict(x) fields listed in [y]
        cnt = 0
        dividends = list(filter(lambda tr: tr['type'] in ['Dividends', 'Payment In Lieu Of Dividends'], cash))
        dividends = [drop_fields(x, ['tid']) for x in dividends]  # remove 'tid' field as not used for dividends
        dividends = self.aggregate_dividends(dividends)
        asset_payments_base = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
        for i, dividend in enumerate(dividends):
            dividend['id'] = asset_payments_base + i
            dividend['type'] = FOF.PAYMENT_DIVIDEND
            self.drop_extra_fields(dividend, ["currency", "reported"])
            self._data[FOF.ASSET_PAYMENTS].append(dividend)
            cnt += 1
        asset_payments_base += cnt
        bond_interests = list(filter(lambda tr: tr['type'] in ['Bond Interest Paid', 'Bond Interest Received'], cash))
        for i, bond_interest in enumerate(bond_interests):
            bond_interest['id'] = asset_payments_base + i
            bond_interest['type'] = FOF.PAYMENT_INTEREST
            self.drop_extra_fields(bond_interest, ["currency", "reported", "tid"])
            self._data[FOF.ASSET_PAYMENTS].append(bond_interest)
            cnt += 1

        taxes = list(filter(lambda tr: tr['type'] == 'Withholding Tax', cash))
        taxes = [drop_fields(x, ['tid']) for x in taxes]
        taxes = self.aggregate_taxes(taxes)
        for tax in taxes:
            cnt += self.apply_tax_withheld(tax)

        transfer_base = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfers = list(filter(lambda tr: tr['type'] == 'Deposits/Withdrawals', cash))
        for i, transfer in enumerate(transfers):
            transfer['id'] = transfer_base + i
            transfer['number'] = transfer.pop('tid')
            transfer['asset'] = [transfer['currency'], transfer['currency']]
            if transfer['amount'] >= 0:  # Deposit
                transfer['account'] = [0, transfer['account'], 0]
                transfer['withdrawal'] = transfer['deposit'] = transfer['amount']
            else:  # Withdrawal
                transfer['account'] = [transfer['account'], 0, 0]
                transfer['withdrawal'] = transfer['deposit'] = -transfer['amount']
            transfer['fee'] = 0.0
            self.drop_extra_fields(transfer, ["type", "amount", "currency", "reported"])
            self._data[FOF.TRANSFERS].append(transfer)
            cnt += 1

        payment_base = max([0] + [x['id'] for x in self._data[FOF.INCOME_SPENDING]]) + 1
        fees = list(filter(lambda tr: 'type' in tr and tr['type'] in ['Other Fees',
                                                                      'Commission Adjustments',  #FIXME Link this fee with asset
                                                                      'Broker Interest Paid',
                                                                      'Broker Interest Received'], cash))
        for i, fee in enumerate(fees):
            fee['id'] = payment_base + i
            fee['peer'] = 0
            if fee['type'] == 'Broker Interest Received':
                category = -PredefinedCategory.Interest
            else:
                category = -PredefinedCategory.Fees
            fee['lines'] = [{'amount': fee['amount'], 'category': category, 'description': fee['description']}]
            self.drop_extra_fields(fee, ["type", "amount", "description", "asset", "number", "currency", "reported", "tid"])
            self._data[FOF.INCOME_SPENDING].append(fee)
            cnt += 1

        logging.info(self.tr("Cash transactions loaded: ") + f"{cnt} ({len(cash)})")

    # Method takes a list of dividend dictionaries and checks for REVERSAL and CANCEL
    # For such description it looks for matching record (for the same symbol) with opposite amount and the same payment
    # and report dates. Description may be different!
    def aggregate_dividends(self, dividends: list) -> list:
        is_reversal = lambda x: self.ReversalSuffix in x or x.startswith(self.CancelPrefix)
        payments = [x for x in deepcopy(dividends) if not is_reversal(x['description'])]
        reversals = [x for x in deepcopy(dividends) if is_reversal(x['description'])]
        # Drop reversals that match payments exactly
        for reversal in reversals:
            t_payment = deepcopy(reversal)  # target payment to search for
            t_payment['description'] = t_payment['description'].replace(self.ReversalSuffix, '')
            t_payment['description'] = t_payment['description'].replace(self.CancelPrefix, '')
            t_payment['amount'] = -t_payment['amount']
            if t_payment not in payments:
                no_description = lambda x: {i:x[i] for i in x if i != 'description'}
                m_payments = [x for x in payments if no_description(x) == no_description(t_payment)]
                if len(m_payments):
                    payments.remove(m_payments[0])
                    logging.warning(self.tr("Payment was reversed by approximate description: ") +
                                    f"{ts2dt(t_payment['timestamp'])}, '{t_payment['description']}': {t_payment['amount']}")
                    continue
                no_reported = lambda x: {i: x[i] for i in x if i != 'reported'}
                m_payments = [x for x in payments if no_reported(x) == no_reported(t_payment)]
                if len(m_payments):
                    payments.remove(m_payments[0])   # FIXME - this branch may lead to non-reversed taxes theoretically
                    logging.warning(self.tr("Payment was reversed with different reported date: ") +
                                    f"{ts2dt(t_payment['timestamp'])}, '{t_payment['description']}': {t_payment['amount']}")
                    continue
                raise Statement_ImportError(self.tr("Can't find match for reversal: ") + f"{reversal}")
            else:
                payments.remove(t_payment) # Source payment found -> remove it
                logging.info(self.tr("Payment was reversed: ") +
                             f"{ts2dt(t_payment['timestamp'])}, '{t_payment['description']}': {t_payment['amount']}")
        return payments

    # Method takes a list of taxes and checks if we have the same amount added and deducted the same day
    # First it tries to find exact match. Second it does it again ignoring reportDate.
    def aggregate_taxes(self, taxes: list) -> list:
        payments = [x for x in deepcopy(taxes) if x['amount'] < 0]
        reversals = [x for x in deepcopy(taxes) if x['amount'] > 0]
        not_matched_reversals = []
        for reversal in reversals:
            t_payment = deepcopy(reversal)   # target payment to search for
            t_payment['description'] = t_payment['description'].replace(self.CancelPrefix, '')
            t_payment['amount'] = -t_payment['amount']
            if t_payment not in payments:
                no_report = lambda x: {i: x[i] for i in x if i != 'reported'}
                m_payments = [x for x in payments if no_report(x) == no_report(t_payment)]
                if len(m_payments):
                    payments.remove(m_payments[0])
                else:
                    no_d_and_r = lambda x: {i: x[i] for i in x if i != 'description' and i != 'reported'}
                    m_payments = [x for x in payments if no_d_and_r(x) == no_d_and_r(t_payment)]
                    if len(m_payments):
                        payments.remove(m_payments[0])
                    else:
                        not_matched_reversals.append(reversal)
            else:
                payments.remove(t_payment)    # it is possible to kill exact match silently
        taxes = [x for x in taxes if x in payments or x in not_matched_reversals]

        # Sometimes IB split tax in several parts for Payment in Lieu of Dividend
        # Below code aggregates such taxes but only negative values (positive might be a correction of previous tax)
        key_func = lambda x: (x['account'], x['asset'], x['currency'], x['description'], x['timestamp'], x['reported'])
        taxes_sorted = sorted(taxes, key=key_func)
        tax_in_lieu = [x for x in taxes_sorted if x['amount'] < 0 and 'PAYMENT IN LIEU OF DIVIDEND' in x['description']]
        other_taxes = [x for x in taxes_sorted if x not in tax_in_lieu]
        lieu_aggregated = []
        for k, group in groupby(tax_in_lieu, key=key_func):
            group_list = list(group)
            part = group_list[0]  # Take fist of several actions as a basis
            part['amount'] = sum(tax['amount'] for tax in group_list)  # and update quantity in it
            lieu_aggregated.append(part)
        taxes_aggregated = sorted(other_taxes + lieu_aggregated, key=key_func)
        return taxes_aggregated

    def load_taxes(self, taxes):
        cnt = 0   #FIXME Link this tax with asset
        tax_base = max([0] + [x['id'] for x in self._data[FOF.INCOME_SPENDING]]) + 1
        for i, tax in enumerate(taxes):
            tax['id'] = tax_base + i
            tax['peer'] = 0
            note = f"{tax['symbol']} ({tax['description']}) - {tax['tax_description']}"
            tax['lines'] = [{'amount': tax['amount'], 'category': -PredefinedCategory.Taxes, 'description': note}]
            self.drop_extra_fields(tax, ["symbol", "amount", "description", "tax_description"])
            self._data[FOF.INCOME_SPENDING].append(tax)
            cnt += 1
        logging.info(self.tr("Taxes loaded: ") + f"{cnt} ({len(taxes)})")

    def load_cfd_charges(self, charges):
        cnt = 0
        charges_base = max([0] + [x['id'] for x in self._data[FOF.INCOME_SPENDING]]) + 1
        for i, charge in enumerate(charges):
            if charge['asset'] != self.NoAsset and not charge['description'].startswith('CFD BORROW FEE FOR'):
                # FIXME if asset is present -> put this charge not in Income/Spending but in Asset Payments section
                logging.warning(self.tr("Unknown CFD charge description: ") + charge['description'])
                continue
            if charge['asset'] == self.NoAsset and not (
                    charge['description'].startswith('LONG CFD INTEREST FOR') or
                    charge['description'].startswith('SHORT CFD INTEREST FOR')):
                logging.warning(self.tr("Unknown CFD charge description: ") + charge['description'])
                continue
            charge['id'] = charges_base + i
            charge['peer'] = 0
            charge['lines'] = [{'amount': charge['amount'], 'category': -PredefinedCategory.Fees, 'description': charge['description']}]
            self.drop_extra_fields(charge, ["amount", "asset", "description", "number"])
            self._data[FOF.INCOME_SPENDING].append(charge)
            cnt += 1
        logging.info(self.tr("CFD charges loaded: ") + f"{cnt} ({len(charges)})")

    # Applies tax to matching dividend:
    # if tax < 0: apply it to dividend without tax
    # otherwise: it is a correction and there should be dividend with exactly the same tax that will be set to 0
    def apply_tax_withheld(self, tax) -> int:
        TaxFullPattern = r"^(?P<description>.*) - (?P<country>\w\w) TAX$"

        parts = re.match(TaxFullPattern, tax['description'], re.IGNORECASE)
        if not parts:
            logging.warning(self.tr("*** MANUAL ENTRY REQUIRED ***"))
            logging.warning(self.tr("Unhandled tax country pattern found: ") + f"{tax['description']}")
            return 0
        parts = parts.groupdict()
        self.set_asset_counry(tax['asset'], parts['country'].lower())
        description = parts['description']
        previous_tax = Decimal(tax['amount']) if Decimal(tax['amount']) >= Decimal('0') else Decimal('0')
        new_tax = -tax['amount'] if tax['amount'] < 0 else 0

        dividend = self.find_dividend4tax(tax['timestamp'], tax['account'], tax['asset'], previous_tax, new_tax, description)
        if dividend is None:
            self.save_debug_info(account=tax['account'], asset=tax['asset'])
            raise Statement_ImportError(self.tr("Dividend not found for withholding tax: ") + f"{tax}, {previous_tax}")
        if dividend['id'] < 0:  # Notification is required if we adjust data for dividend that is already in Jal DB
            logging.info(self.tr("Tax adjustment for dividend: ") +
                         f"{dividend['tax']} -> {new_tax} ({ts2dt(dividend['timestamp'])} {dividend['description']})")
        dividend["tax"] = new_tax
        # append new dividend if it came from DB and haven't been loaded in self._data yet
        if len([1 for x in self._data[FOF.ASSET_PAYMENTS] if x['id'] == dividend['id']]) == 0:
            dividend['type'] = FOF.PAYMENT_DIVIDEND
            self._data[FOF.ASSET_PAYMENTS].append(dividend)
        return 1

    # Searches for dividend that matches tax in the best way:
    # - it should have exactly the same account_id and asset_id
    # - tax amount withheld from dividend should be equal to provided 'tax' value
    # - timestamp should be the same or within previous year for weak match of Q1 taxes
    # - note should be exactly the same or contain the same key elements
    def find_dividend4tax(self, timestamp, account_id, asset_id, prev_tax, new_tax, note):
        PaymentInLiueOfDividend = 'PAYMENT IN LIEU OF DIVIDEND'
        TaxNotePattern = r"^(?P<symbol>.*\w) ?\((?P<isin>\w+)\)(?P<prefix>( \w*)+) +(?P<amount>\d+\.\d+)?(?P<suffix>.*)$"
        DividendNotePattern = r"^(?P<symbol>.*\w) ?\((?P<isin>\w+)\)(?P<prefix>( \w*)+) +(?P<amount>\d+\.\d+)?(?P<suffix>.*) \(.*\)$"

        dividends = [x for x in self._data[FOF.ASSET_PAYMENTS] if
                     (x['type'] == FOF.PAYMENT_DIVIDEND or x['type'] == FOF.PAYMENT_STOCK_DIVIDEND)
                     and x['asset'] == asset_id and x['account'] == account_id]
        db_account = self._map_db_account(account_id)
        db_asset = self._map_db_asset(asset_id)
        if db_account and db_asset:
            for db_dividend in Dividend.get_list(db_account, db_asset, Dividend.Dividend):
                dividends.append({
                    "id": -db_dividend.oid(),
                    "account": account_id,
                    "asset": asset_id,
                    "timestamp": db_dividend.timestamp(),
                    "number": db_dividend.number(),
                    "amount": float(db_dividend.amount()),
                    "tax": float(db_dividend.tax()),
                    "description": db_dividend.note()
                })
        if datetime.utcfromtimestamp(timestamp).timetuple().tm_yday < 75:
            # We may have wrong date in taxes before March, 15 due to tax correction
            range_start, _range_end = ManipulateDate.PreviousYear(day=datetime.utcfromtimestamp(timestamp))
            dividends = [x for x in dividends if x['timestamp'] >= range_start]
        else:
            # For any other day - use exact time match
            dividends = [x for x in dividends if x['timestamp'] == timestamp]
        dividends = [x for x in dividends if 'tax' not in x or (abs(Decimal(x['tax']) - prev_tax) < 0.0001)]
        dividends = sorted(dividends, key=lambda x: x['timestamp'])

        # Choose either Dividends or Payments in liue with regards to note of the matching tax
        if PaymentInLiueOfDividend in note.upper():
            dividends = list(filter(lambda item: PaymentInLiueOfDividend in item['description'], dividends))
            # we don't check for full match as there are a lot of records without amount
        else:
            dividends = list(filter(lambda item: PaymentInLiueOfDividend not in item['description'], dividends))
            # Check for full match
            for dividend in dividends:
                if (dividend['timestamp'] == timestamp) and (note.upper() == dividend['description'][:len(note)].upper()):
                    return dividend
        if len(dividends) == 0:
            return None

        # Chose most probable dividend - by amount, timestamp and description
        parts = re.match(TaxNotePattern, note, re.IGNORECASE)
        if not parts:
            logging.warning(self.tr("*** MANUAL ENTRY REQUIRED ***"))
            logging.warning(self.tr("Unhandled tax pattern found: ") + f"{note}")
            return None
        parts = parts.groupdict()
        note_prefix = parts['prefix']
        note_suffix = parts['suffix']
        try:
            note_amount = float(parts['amount'])
        except (ValueError, TypeError):
            note_amount = 0
        score = [0] * len(dividends)
        for i, dividend in enumerate(dividends):
            parts = re.match(DividendNotePattern, dividend['description'], re.IGNORECASE)
            if not parts:
                logging.warning(self.tr("*** MANUAL ENTRY REQUIRED ***"))
                logging.warning(self.tr("Unhandled dividend pattern found: ") + f"{dividend['description']}")
                return None
            parts = parts.groupdict()
            try:
                amount = float(parts['amount'])
            except (ValueError, TypeError):
                amount = 0
            if abs(amount - note_amount) <= 0.000005:            # Description has very similar amount +++++
                score[i] += 5
            if dividend['timestamp'] == timestamp:               # Timestamp exact match gives ++
                score[i] += 2
            if abs(0.1 * dividend['amount'] - new_tax) <= 0.01:  # New tax is 10% of dividend gives +
                score[i] += 1
            if parts['prefix'] == note_prefix:                   # Prefix part of description match gives +
                score[i] += 1
            if parts['suffix'] == note_suffix:                   # Suffix part of description match gives +
                score[i] += 1
        for i, vote in enumerate(score):
            if (vote == max(score)) and (vote > 0):
                return dividends[i]
        # Final check - if only one found, return it
        if len(dividends) == 1:
            return dividends[0]
        return None

    # Assign ex-date from dividend accruals
    def load_dividend_accruals(self, accruals):
        posted = [x for x in accruals if x['code'] == StatementIBKR.ReversalCode]
        for accrual in posted:
            dividends = [x for x in self._data[FOF.ASSET_PAYMENTS] if x['account']==accrual['account'] and x['asset']==accrual['asset'] and ts2d(x['timestamp'])==ts2d(accrual['timestamp']) and x['amount']==-accrual['amount']]
            if len(dividends) == 1:
                dividend = dividends[0]
            else:
                continue
            dividend['ex_date'] = accrual['ex_date']

    # Removes data that was used during XML processing but isn't needed in final output:
    # Drop any assets with type 'right' as JAL won't import them
    def strip_unused_data(self):
        rights_id = [x['id'] for x in self._data[FOF.ASSETS] if x['type'] == FOF.ASSET_RIGHTS]
        for asset_id in rights_id:
            self.remove_asset(asset_id)

    # -----------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def order_statements(statementFiles) -> list:
        READ_COUNT=1024
        pattern = re.compile(r'<FlexStatement.*fromDate=\"([0-9]*)\"\stoDate=\"([0-9]*)\".*>')
        files = []
        for file in statementFiles:
            with open(file) as f:
                m = re.search(pattern, f.read(READ_COUNT))
                if not m:
                    logging.error(StatementIBKR.tr("Can't find a FlexStatement in first {} bytes of {}").format(READ_COUNT,file))
                    return []
                start = int(m.group(1))
                end = int(m.group(2))
                item = [start, end, file]
                idx = 0
                for i in files:
                    if item[1] <= i[0]:
                        break
                    idx += 1
                files.insert(idx, item)
                f.close()
        return [x[2] for x in files]
